###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import pytest

import LbAPCommon


@pytest.mark.parametrize(
    "app_name,app_version,expected",
    [
        ("DaVinci", "v36r3p1", True),
        ("DaVinci", "v36r4p1", False),
        ("davinci", "v36r3p1", True),
        ("davinci", "v36r4p1", False),
        ("DAVINCI", "v36r3p1", True),
        ("DAVINCI", "v36r4p1", False),
        ("DaVinci", "v3", True),
        ("DaVinci", "v35r33", True),
        ("DaVinci", "v35r3p1111", True),
        ("DAVINCI", "v50", False),
        ("DAVINCI", "v50r4", False),
        ("DAVINCI", "v50r4p11", False),
        ("Castelo", "v3r0", False),
        ("Castelo", "not-a-version", False),
    ],
)
def test_project_uses_cmt(app_name, app_version, expected):
    assert LbAPCommon.hacks.project_uses_cmt(app_name, app_version) is expected



###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
"""Contains utility functions used to display and save the output of the checks.
"""
import copy
import json
import warnings

import hist
import numpy as np
import uproot
from hist import Hist

from LbAPCommon.checks import CheckResult


def hist_to_root(job_name, check_results, output_path):
    """Save histograms to a root file.

    Creates a .root file at the provided output location, containing any
    histograms stored in the check results.

    Args:
        job_name (str): Name of the job within the production for which these
          checks were run. Unused.
        check_results (dict): Results of all checks performed for that job, as
          returned by checks.run_job_checks().
        output_path (str): Path of the root file to be created
    """
    # Create the file only if the check produce histograms in output
    checks_with_histo = ["range", "range_nd", "range_bkg_subtracted"]
    with uproot.recreate(output_path) as file_root:
        for cr in check_results:
            if (
                check_results[cr].passed
                and check_results[cr].check_type in checks_with_histo
            ):
                for key, data in check_results[cr].tree_data.items():
                    for hist_counter, _histo in enumerate(data.get("histograms", [])):
                        histo_name = f"{key}/{cr}_{hist_counter}"
                        file_root[histo_name] = _histo


def serialise_hist(obj):
    """Converts histograms into JSON.

    Creates a JSON serialisation of a histogram from the hist library. This
    uses a custom serialisation format, defined by this function.

    Args:
        obj (hist.basehist.BaseHist): A histogram from the hist library.

    Returns:
        JSON-style dict containing hist information (version 1).
    """
    if isinstance(obj.variances(flow=True), type(None)):
        raise ValueError(
            "Filling an unweighted storage with weights will result in incorrectly calculated variances"
        )

    serialised_obj = {
        "version": 1,
        "name": str(obj.name),
        "axes": [
            {
                "name": str(axis.name),
                "nbins": int(len(axis)),
                "min": float(axis.edges[0]),
                "max": float(axis.edges[-1]),
            }
            for axis in obj.axes
        ],
        "contents": list(obj.values(flow=True).tolist()),
        "sumw2": list(obj.variances(flow=True).tolist()),
    }
    return serialised_obj


def deserialise_hist(hist_json, check_type):
    """Converts custom JSON representation of histograms into a hist.

    Creates a histogram from the hist library from serialised JSON, as defined
    by serialise_hist(). This is made backwards-compatible using the "version"
    property of the serialised JSON, which allows hists to be reconstructed
    differently depending on the version number.

    Args:
        hist_json (dict): JSON-style dict containing hist serialisation.

    Returns:
        hist library histogram
    """
    if hist_json["version"] == 1:
        axes = [
            hist.axis.Regular(ax["nbins"], ax["min"], ax["max"], name=ax["name"])
            for ax in hist_json["axes"]
        ]
        hist_storage = (
            hist.storage.Double()
            if check_type != "range_bkg_subtracted"
            else hist.storage.Weight()
        )
        if 1 <= len(axes) <= 3:
            h = Hist(*axes, name=hist_json["name"], storage=hist_storage)
        else:
            raise ValueError(f"Expected between 1 and 3 axes, got {len(axes)}")

        # fill hists with values & variances
        if check_type == "range_bkg_subtracted":
            h.view(flow=True).value = np.array(hist_json["contents"])
            h.view(flow=True).variance = np.array(hist_json["sumw2"])
        else:
            h.view(flow=True)[:] = np.array(hist_json["contents"])
            # for Double() storage hists, variances are set automatically by this

        return h
    else:
        raise ValueError(
            f"Invalid hist serialisation version ({hist_json['version']}) selected!"
        )


def checks_to_JSON(
    checks_data,
    all_check_results,
    json_output_path=None,
):
    """Converts information about all checks into JSON.

    Args:
        checks_data (dict): Information defining checks, as returned by
          parsing.parse_yaml()
        all_check_results (dict): dict of results of checks for all different
          jobs. For one job, = {"job_name": checks.run_job_checks(...)}.
        json_output_path (str, None): If not None, will additionally write the
          output JSON to a file at the provided location.

    Returns:
        Plaintext JSON (string) containing check information
    """
    all_check_results_copy = copy.deepcopy(all_check_results)

    result = {}
    for job in all_check_results_copy:
        result[job] = {}
        for check in all_check_results_copy[job]:
            result[job][check] = {}

            result[job][check]["passed"] = all_check_results_copy[job][check].passed
            result[job][check]["messages"] = all_check_results_copy[job][check].messages
            result[job][check]["can_combine"] = all_check_results_copy[job][
                check
            ].can_combine
            result[job][check]["input"] = checks_data[check]
            result[job][check]["output"] = all_check_results_copy[job][check].tree_data

            # Convert histograms to JSON representation
            for tree in result[job][check]["output"]:
                if "histograms" in result[job][check]["output"][tree]:
                    n_hists = len(result[job][check]["output"][tree]["histograms"])
                    for n in range(n_hists):
                        result[job][check]["output"][tree]["histograms"][
                            n
                        ] = serialise_hist(
                            result[job][check]["output"][tree]["histograms"][n]
                        )

    if json_output_path is not None:
        with open(json_output_path, "w", encoding="utf8") as json_file:
            json.dump(result, json_file, indent="  ")

    return json.dumps(result, indent="  ")


def JSON_to_checks(check_results_json):
    """Converts JSON of check results into check data and results.

    Deserialise information about all checks from a JSON format back into check
    data and results.

    Args:
        check_results_json (dict): JSON-style dict containing serialised check
          information, as created by checks_to_JSON().

    Returns:
        A tuple (checks_data, all_check_results) where checks_data contains
          information on how the checks are defined, and all_check_results
          contains the results of all checks. These return values should match
          exactly what was passed to checks_to_JSON() when creating this JSON.
    """
    checks_data = {}
    all_check_results = {}

    for job in check_results_json:
        all_check_results[job] = {}
        for check, json_data in check_results_json[job].items():

            checks_data[check] = json_data["input"]

            all_check_results[job][check] = CheckResult(
                check_type=json_data["input"]["type"],
                passed=json_data["passed"],
                messages=json_data["messages"],
                can_combine=json_data["can_combine"],
                tree_data=json_data["output"],
            )

            # Convert histograms from JSON representation
            for tree in all_check_results[job][check].tree_data:
                if "histograms" in all_check_results[job][check].tree_data[tree]:
                    for n, h in enumerate(
                        all_check_results[job][check].tree_data[tree]["histograms"]
                    ):
                        all_check_results[job][check].tree_data[tree]["histograms"][
                            n
                        ] = deserialise_hist(
                            h, all_check_results[job][check].check_type
                        )

    return checks_data, all_check_results


def _add_check_results(check_data, *results):
    """Add results from multiple of the same check on separate data.

    Adds together the results of the same check on multiple different datasets.
    All sets of results should be from the same job on different data, and
    should also be defined by identical YAML config. This function does not
    check that these conditions are met, this is done in combine_checks().

    Args:
        check_data (dict): YAML-type dictionary which configures the check.
        *results: Variable length argument list. All arguments should be of
          type CheckResult, and be created by the same check.

    Returns:
        A CheckResult containing the combination of results. This should
        be equivalent to if the check had been run once on all datasets
        combined (with the exception of the messages).
    """
    # Input validation
    if not all([isinstance(res, CheckResult) for res in results]):
        raise ValueError(
            "Arguments passed to *results must all be of type CheckResult "
            f"(result types: {[type(res) for res in results]})"
        )
    if not all([results[0].check_type == res.check_type for res in results]):
        raise ValueError(
            "Arguments passed to *results must have the same check type "
            f"(result check types: {[res.check_type for res in results]})"
        )
    if not all(
        [
            list(results[0].tree_data.keys()) == list(res.tree_data.keys())
            for res in results
        ]
    ):
        raise ValueError(
            "Arguments passed to *results must have the same list of TTrees in tree_data "
            f"(result TTrees: {[list(res.tree_data.keys()) for res in results]})"
        )
    if len(results) == 0:
        raise ValueError("Found 0 checks to be combined")

    # Create basic CheckResult with everything except tree_data combined
    added_result = CheckResult(
        check_type=results[0].check_type,
        passed=True,  # this gets modified later by the check result addition
        messages=[msg for res in results for msg in res.messages],
        can_combine=True,  # the output of a combination can be combined again with other CheckResults objects if necessary
        tree_data={tree: {} for tree in results[0].tree_data},
    )

    # Add data together
    for tree in results[0].tree_data:
        data_added = added_result.tree_data[tree]

        # range check
        if added_result.check_type == "range":
            data_added["histograms"] = []
            data_added["histograms"].append(
                sum([res.tree_data[tree]["histograms"][0] for res in results])
            )
            data_added["num_entries"] = data_added["histograms"][0].sum()
            bin_centers = data_added["histograms"][0].axes[0].centers
            data_added["mean"] = (
                sum(bin_centers * data_added["histograms"][0].values())
                / data_added["histograms"][0].sum()
            )
            if data_added["num_entries"] >= 2:
                data_added["variance"] = sum(
                    (bin_centers - data_added["mean"]) ** 2
                    * data_added["histograms"][0].values()
                ) / (data_added["histograms"][0].sum() - 1)
            else:
                data_added["variance"] = 0
            data_added["stddev"] = data_added["variance"] ** 0.5
            data_added["num_entries_in_mean_window"] = sum(
                [res.tree_data[tree]["num_entries_in_mean_window"] for res in results]
            )
            # update check result
            added_result.passed = added_result.passed and all(
                [
                    data_added["num_entries"] > 0,
                    abs(data_added["mean"] - check_data["exp_mean"])
                    <= check_data["mean_tolerance"]
                    if check_data.get("exp_mean") is not None
                    and check_data.get("mean_tolerance") is not None
                    else True,
                    abs(data_added["stddev"] - check_data["exp_std"])
                    <= check_data["std_tolerance"]
                    if check_data.get("exp_std") is not None
                    and check_data.get("std_tolerance") is not None
                    else True,
                ]
            )

        # range_bkg_subtracted check
        elif added_result.check_type == "range_bkg_subtracted":
            data_added["histograms"] = []
            for i in range(4):
                data_added["histograms"].append(
                    sum([res.tree_data[tree]["histograms"][i] for res in results])
                )
            # update check result
            added_result.passed = added_result.passed and all(
                [
                    data_added["histograms"][1].view().value.sum() > 0,
                    data_added["histograms"][2].view().value.sum() > 0,
                ]
            )

        # range_nd check
        elif added_result.check_type == "range_nd":
            data_added["histograms"] = []
            data_added["histograms"].append(
                sum([res.tree_data[tree]["histograms"][0] for res in results])
            )
            data_added["num_entries"] = data_added["histograms"][0].sum()
            # update check result
            added_result.passed = added_result.passed and (
                data_added["num_entries"] > 0
            )

        # num_entries check
        elif added_result.check_type == "num_entries":
            data_added["num_entries"] = sum(
                [res.tree_data[tree]["num_entries"] for res in results]
            )
            # update check result
            added_result.passed = added_result.passed and (
                data_added["num_entries"] >= check_data["count"]
            )

        # num_entries_per_invpb check
        elif added_result.check_type == "num_entries_per_invpb":
            data_added["num_entries"] = sum(
                [res.tree_data[tree]["num_entries"] for res in results]
            )
            data_added["lumi_invpb"] = sum(
                [res.tree_data[tree]["lumi_invpb"] for res in results]
            )
            data_added["num_entries_per_invpb"] = (
                data_added["num_entries"] / data_added["lumi_invpb"]
            )
            # update check result
            added_result.passed = added_result.passed and (
                data_added["num_entries_per_invpb"] >= check_data["count_per_invpb"]
            )

        elif added_result.check_type == "branches_exist":
            # Nothing stored in tree_data for branches_exist checks
            # update check result
            added_result.passed = added_result.passed and all(
                res.passed for res in results
            )

    return added_result


def combine_checks(checks_info_tuple_list):
    """Combine the results of checks on separate datasets.

    Combines check results for separate datasets. In order for results to be
    combined, the input must be identical (names and contents), and the job
    name must match.

    Args:
        checks_info_tuple_list (list): List containing tuples of check info
          (checks_data, all_check_results), as described in the args of
          checks_to_JSON(). Each element of the list should correspond to a
          different dataset so there is no duplication when combining.

    Returns:
        A tuple (checks_data, comb_check_results) where checks_data contains
          information on how the checks are defined, and comb_check_results
          contains the results of all checks after combining.
    """
    # Input validation: checks_info_tuple_list should be a list of tuples (len=2) of dicts
    if not all(len(tup) == 2 for tup in checks_info_tuple_list):
        raise ValueError(
            "Invalid checks_info_tuple_list passed - all elements must be tuples of length 2"
        )

    # Establish the combined checks_data with info from user's production YAML
    checks_data = {}
    for tup in checks_info_tuple_list:
        data, _results = tup
        for c_name, c_data in data.items():
            # Add check if not already present
            if c_name not in checks_data:
                checks_data[c_name] = c_data
            # Otherwise, confirm that the check is configured identically
            else:
                if c_data != checks_data[c_name]:
                    warnings.warn(
                        f"Found check with duplicate name ({c_name}) but different configuration - cannot be combined"
                    )

    # Create a dict containing indices of datasets to include when combining a check for each job
    checks_datasets = {}
    for i, tup in enumerate(checks_info_tuple_list):
        data, results = tup
        for job_name, job_check_results in results.items():

            # Add job if not already present
            if job_name not in checks_datasets:
                checks_datasets[job_name] = {}

            # Fill in dataset list
            for check_name in job_check_results:
                # Skip if this is a duplicate check with different configuration
                if checks_data[check_name] != data[check_name]:
                    continue

                # Add check to that job if not already present
                if check_name not in checks_datasets[job_name]:
                    checks_datasets[job_name][check_name] = []
                # Don't combine checks if the can_combine flag is set to False.
                if not job_check_results[check_name].can_combine:
                    warnings.warn(
                        f"The result for {job_name} cannot be included in the {check_name} "
                        "combination as the can_combine flag is False."
                    )
                    continue
                checks_datasets[job_name][check_name].append(i)

    # Combine results
    comb_check_results = {}
    for job_name in checks_datasets:
        comb_check_results[job_name] = {}
        for check_name in checks_datasets[job_name]:
            datasets = checks_datasets[job_name][check_name]
            comb_check_results[job_name][check_name] = _add_check_results(
                checks_data[check_name],
                *[
                    cit[1][job_name][check_name]
                    for i, cit in enumerate(checks_info_tuple_list)
                    if i in datasets
                ],
            )

    return checks_data, comb_check_results



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import pytest

from LbAPCommon.parsing import create_proc_pass_map


@pytest.mark.parametrize(
    "jobs, expected",
    [
        [
            ["MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_tuple"],
            {
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_tuple": "AnaProd-v0r0p00000000-MC_Bu2JpsiKplus_no_str_g_Sim09g_tuple"
            },
        ],
        [["______my-job"], {"______my-job": "AnaProd-v0r0p00000000-my-job"}],
        [["my---job"], {"my---job": "AnaProd-v0r0p00000000-my-job"}],
        [
            [
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_strip",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_tuple",
            ],
            {
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_strip": "AnaProd-v0r0p00000000-MC_Bu2JpsiKplus_no_str_g_Sim09g_strip",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_tuple": "tuple",
            },
        ],
        [
            [
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_tuple",
            ],
            {
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g": "AnaProd-v0r0p00000000-MC_Bu2JpsiKplus_no_str_g_Sim09g",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_tuple": "tuple",
            },
        ],
        [
            [
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_strip",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g",
            ],
            {
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_strip": "AnaProd-v0r0p00000000-MC_Bu2JpsiKplus_no_str_g_Sim09g_strip",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g": "default",
            },
        ],
        [
            [
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_hlt",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_strip",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_tuple",
            ],
            {
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_hlt": "AnaProd-v0r0p00000000-MC_Bu2JpsiKplus_no_str_g_Sim09g_hlt",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_strip": "strip",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_tuple": "tuple",
            },
        ],
        [
            [
                "MC_2017_MagUp_Bu2JpsiKplus_Sim09g_strip_hlt",
                "MC_2017_MagUp_Bu2JpsiKplus_Sim09g_strip",
                "MC_2017_MagUp_Bu2JpsiKplus_Sim09g",
            ],
            {
                "MC_2017_MagUp_Bu2JpsiKplus_Sim09g_strip_hlt": "AnaProd-v0r0p00000000-MC_Bu2JpsiKplus_Sim09g_strip_hlt",
                "MC_2017_MagUp_Bu2JpsiKplus_Sim09g_strip": "default",
                "MC_2017_MagUp_Bu2JpsiKplus_Sim09g": "default",
            },
        ],
        [
            [
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_hlt",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_hlt+strip",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_hlt+strip+tuple",
            ],
            {
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_hlt": "AnaProd-v0r0p00000000-MC_Bu2JpsiKplus_no_str_g_Sim09g_hlt",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_hlt+strip": "strip",
                "MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g_hlt+strip+tuple": "tuple",
            },
        ],
    ],
)
def test_create_proc_pass_map_good(jobs, expected):
    proc_pass_map = create_proc_pass_map(jobs, "v0r0p00000000")
    assert proc_pass_map == expected


@pytest.mark.parametrize(
    "jobs",
    [
        ["A" * 1000],
        [
            "hlt_MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g",
            "strip_MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g",
            "tuple_MC_2017_MagUp_Bu2JpsiKplus_no_str_g_Sim09g",
        ],
    ],
)
def test_create_proc_pass_map_bad(jobs):
    with pytest.raises(ValueError, match=r".*is too long.*"):
        create_proc_pass_map(jobs, "v0r0p00000000")



###############################################################################
# (c) Copyright 2020-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

__all__ = [
    "get_api_token",
    "wipe_token_cache",
    "hacks",
    "render_yaml",
    "parse_yaml",
    "validate_yaml",
    "lint_all",
    "validators",
    "write_jsroot_compression_options",
]

from os.path import join

from . import hacks, validators
from .linting import lint_all
from .parsing import parse_yaml, render_yaml, validate_yaml


def write_jsroot_compression_options(dynamic_dir):
    """Write options file to configure JSROOT-compatible compression on job output files.

    Args:
        dynamic_dir: Location to write the use-jsroot-compression.py options file.
    """
    with open(join(dynamic_dir, "use-jsroot-compression.py"), "wt") as fp:
        fp.write(
            "\n".join(
                [
                    "from Configurables import RootCnvSvc",
                    "RootCnvSvc().GlobalCompression = 'ZLIB:1'",
                    "",
                    "try:",
                    "    from Configurables import DaVinci",
                    "except ImportError:",
                    "    pass",
                    "else:",
                    "    try:",
                    "        DaVinci().RootCompressionLevel = 'ZLIB:1'",
                    "    except AttributeError:",
                    "        pass",
                ]
            )
        )



###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import pytest

from LbAPCommon.linting.bk_paths import validate_bk_query

EXAMPLE_PATH_1 = "/MC/2018/Beam6500GeV-2018-MagDown/Sim09g/Trig0x617d18a4/Reco18/24142001/ALLSTREAMS.DST"


@pytest.mark.parametrize(
    "path",
    [
        EXAMPLE_PATH_1,
        "/LHCb/Collision15/Beam6500GeV-VeloClosed-MagDown/Real Data/Reco15a/Stripping24r1/90000000/MDST.DST",
    ],
)
def test_good(path):
    assert validate_bk_query(EXAMPLE_PATH_1)


def test_suggestion():
    with pytest.raises(ValueError) as e:
        validate_bk_query(
            "/MC/2018/24142001/Beam6500GeV-2018-MagDown/Sim09g/Trig0x617d18a4/Reco18/ALLSTREAMS.DST"
        )
    assert "The provided BK query is not valid" in str(e)
    assert EXAMPLE_PATH_1 in str(e)


@pytest.mark.parametrize(
    "path",
    [
        "/MC/2018/xxxxx/Beam6500GeV-2018-MagDown/Sim09g/Trig0x617d18a4/Reco18/ALLSTREAMS.DST",
    ],
)
def test_invalid(path):
    with pytest.raises(ValueError) as e:
        validate_bk_query(path)
    assert "The provided BK query is not valid" in str(e)



###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import math
import re
from dataclasses import dataclass, field
from itertools import combinations
from typing import Dict, List

import hist
import numpy
import uproot
from hist import Hist


@dataclass
class CheckResult:
    """Class for representing the return result of ntuple checks."""

    check_type: str
    passed: bool
    can_combine: bool
    messages: List[str] = field(default_factory=list)
    tree_data: Dict[str, Dict] = field(default_factory=dict)


def run_job_checks(checks_list, check_data, test_ntuple_path_list):
    """Run the checks listed using data from ntuples.

    Run a list of checks (can be a subset of all the checks defined in check_data).
    Requires that the data ntuples are already created (eg. by `lb-ap test`).

    Returns:
        dict: CheckResult objects
    """
    check_results = {}

    for check in checks_list:
        data = check_data[check]

        if data["type"] == "range":
            check_results[check] = range_check(
                test_ntuple_path_list,
                expression=data.get("expression"),
                limits=data.get("limits"),
                n_bins=data.get("n_bins"),
                blind_ranges=data.get("blind_ranges"),
                exp_mean=data.get("exp_mean"),
                exp_std=data.get("exp_std"),
                mean_tolerance=data.get("mean_tolerance"),
                std_tolerance=data.get("std_tolerance"),
                tree_pattern=data.get("tree_pattern"),
            )
        elif data["type"] == "range_bkg_subtracted":
            check_results[check] = range_check_bkg_subtracted(
                test_ntuple_path_list,
                expression=data.get("expression"),
                limits=data.get("limits"),
                expr_for_subtraction=data.get("expr_for_subtraction"),
                mean_sig=data.get("mean_sig"),
                background_shift=data.get("background_shift"),
                background_window=data.get("background_window"),
                signal_window=data.get("signal_window"),
                n_bins=data.get("n_bins"),
                blind_ranges=data.get("blind_ranges"),
                tree_pattern=data.get("tree_pattern"),
            )
        elif data["type"] == "range_nd":
            check_results[check] = range_check_nd(
                test_ntuple_path_list,
                expressions=data.get("expressions"),
                limits=data.get("limits"),
                n_bins=data.get("n_bins"),
                blind_ranges=data.get("blind_ranges"),
                tree_pattern=data.get("tree_pattern"),
            )
        elif data["type"] == "num_entries":
            check_results[check] = num_entries(
                test_ntuple_path_list,
                count=data.get("count"),
                tree_pattern=data.get("tree_pattern"),
            )
        elif data["type"] == "num_entries_per_invpb":
            check_results[check] = num_entries_per_invpb(
                test_ntuple_path_list,
                count_per_invpb=data.get("count_per_invpb"),
                tree_pattern=data.get("tree_pattern"),
                lumi_pattern=data.get("lumi_pattern"),
            )
        elif data["type"] == "branches_exist":
            check_results[check] = branches_exist(
                test_ntuple_path_list,
                branches=data.get("branches"),
                tree_pattern=data.get("tree_pattern"),
            )

    return check_results


def num_entries(
    filepath_list,
    count,
    tree_pattern,
):
    """Number of entries check.

    Check that all matching TTree objects contain a minimum number of entries.

    Args:
        filepath_list (list[file-like]): List of paths to files to analyse
        count (int): The minimum number of entries required
        tree_pattern (regex): A regular expression for the TTree objects to check

    Returns:
        A CheckResult object, which for each tree contains tree_data key/values:
            num_entries: The total number of events in the TTree
    """
    result = CheckResult("num_entries", False, True)
    for filepath in filepath_list:
        trees_opened = []
        with uproot.open(filepath) as f:
            for key, obj in f.items(cycle=False):
                if not isinstance(obj, uproot.TTree):
                    continue
                if not re.fullmatch(tree_pattern, key):
                    continue
                if key in trees_opened:
                    continue
                trees_opened.append(key)

                # First time: initialise the CheckResult
                if key not in result.tree_data:
                    result.tree_data[key] = {}
                    result.tree_data[key]["num_entries"] = 0

                result.tree_data[key]["num_entries"] += obj.num_entries

    result.passed = (
        all([data["num_entries"] >= count for data in result.tree_data.values()])
        and len(result.tree_data) > 0
    )
    for key, data in result.tree_data.items():
        nentries = data["num_entries"]
        result.messages += [f"Found {nentries} in {key} ({count} required)"]

    # If no matches were found the check should be marked as failed
    if len(result.tree_data) == 0:
        result.passed = False
        result.can_combine = False
        result.messages += [f"No TTree objects found that match {tree_pattern}"]

    return result


def range_check(
    filepath_list,
    expression,
    limits,
    n_bins,
    blind_ranges,
    tree_pattern,
    exp_mean,
    exp_std,
    mean_tolerance,
    std_tolerance,
):
    """Range check.

    Check if there is at least one entry in the TTree object with a specific variable falling in a pre-defined range.
    If the expected mean and standard deviation values are given in input, they are compared with the observed ones
    and their agreement within the provided *_tolerance is checked. It is also possible to blind some regions.

    Args:
        filepath_list (list[file-like]): List of paths to files to analyse
        expression (str): Name of the variable (or expression depending on varibales in the TTree) to be checked
        limits (dict): Pre-defined range for x axis
        n_bins (int): Number of bins for the histogram
        blind_ranges (dict): Regions to be blinded in the histogram
        exp_mean (float): Expected mean value (optional)
        exp_std (float): Expected standard deviation (optional)
        mean_tolerance (float): Maximum shift tolerated between expected and observed mean values (optional)
        std_tolerance (float): Maximum shift tolerated between expected and observed values of standard deviation
            (optional)
        tree_pattern (regex): A regular expression for the TTree object to check

    Returns:
        A CheckResult object, which for each tree contains tree_data key/values:
            histograms (list[Hist]): Filled 1D histogram of the quantity defined by the expression parameter
            num_entries (float): The total number of entries in the histogram (with blind ranges applied)
            mean (float): The mean of the histogram (approximated using binned data)
            variance (float): The variance of the histogram (approximated using binned data)
            stddev (float): The standard deviation of the histogram (approximated using binned data)
            num_entries_in_mean_window (float): The number of events falling in the exp_mean +- exp_std region (with
                blind ranges applied)
    """
    result = CheckResult("range", True, True)
    bin_centers = None

    for filepath in filepath_list:
        trees_opened = []
        with uproot.open(filepath) as f:
            for key, obj in f.items(cycle=False):
                if not isinstance(obj, uproot.TTree):
                    continue
                if not re.fullmatch(tree_pattern, key):
                    continue
                if key in trees_opened:
                    continue
                trees_opened.append(key)

                # First time: initialise the CheckResult
                if key not in result.tree_data:
                    result.tree_data[key] = {}
                    axis0 = hist.axis.Regular(
                        n_bins, limits["min"], limits["max"], name=expression
                    )
                    bin_centers = axis0.centers
                    h = Hist(axis0, name=f"{key} {expression}")
                    result.tree_data[key]["histograms"] = [h]
                    result.tree_data[key]["num_entries"] = 0
                    result.tree_data[key]["mean"] = 0
                    result.tree_data[key]["variance"] = 0
                    result.tree_data[key]["stddev"] = 0
                    result.tree_data[key]["num_entries_in_mean_window"] = 0

                values_obj = {}
                # Check if the branch is in the Tree or if the expression is correctly written
                try:
                    values_obj = obj.arrays(expression, library="np")
                except uproot.exceptions.KeyInFileError as e:
                    result.messages += [f"Missing branch in {key!r} with {e!r}"]
                    result.passed = False
                    result.can_combine = False
                    continue

                # Evaluate expression & apply limits
                test_array = values_obj[expression]
                test_array = test_array[
                    numpy.where(
                        (test_array < limits["max"]) & (test_array > limits["min"])
                    )
                ]

                # Apply blinding
                if blind_ranges is not None:
                    if isinstance(blind_ranges, dict):
                        # Take into account that there could be multiple regions to blind
                        blind_ranges = [blind_ranges]
                    for blind_range in blind_ranges:
                        lower, upper = blind_range["min"], blind_range["max"]
                        test_array = test_array[
                            ~((lower < test_array) & (test_array < upper))
                        ]

                # Fill histogram
                result.tree_data[key]["histograms"][0].fill(test_array)

                # Add to event counters
                result.tree_data[key]["num_entries"] += test_array.size

                if exp_mean is not None and exp_std is not None:
                    events_in_exp_mean_region = test_array[
                        (exp_mean - exp_std < test_array)
                        & (test_array < exp_mean + exp_std)
                    ]
                    result.tree_data[key][
                        "num_entries_in_mean_window"
                    ] += events_in_exp_mean_region.size

    # If no matches are found the check should be marked as failed, and can return early
    if len(result.tree_data) == 0:
        result.passed = False
        result.can_combine = False
        result.messages += [f"No TTree objects found that match {tree_pattern}"]
        return result

    # Check the completely filled histograms
    if result.passed:
        for key in result.tree_data:
            # Get this tree's histogram
            h = result.tree_data[key]["histograms"][0]

            # Require at least one event
            # However, even if the histogram is empty, we want to be able to combine
            # the result later (so "can_combine" flag is left "True")
            if h.sum() == 0:
                result.passed = False
                result.messages += [f"No events found in range for Tree {key}"]
                continue

            # Calculate mean, variance, & standard deviation
            mean = sum(bin_centers * h.values()) / h.sum()
            result.tree_data[key]["mean"] = mean

            if h.sum() >= 2:
                variance = sum((bin_centers - mean) ** 2 * h.values()) / (h.sum() - 1)
            else:
                variance = 0
            result.tree_data[key]["variance"] = variance

            stddev = variance**0.5
            result.tree_data[key]["stddev"] = stddev

            # Apply expected mean requirement
            if exp_mean is not None and mean_tolerance is not None:
                delta_mean = abs(mean - exp_mean)
                if delta_mean > mean_tolerance:
                    result.passed = False
                    result.messages += [
                        f"The observed mean ({mean}) differs from the expected value by {delta_mean} (<={mean_tolerance} required)"
                    ]
                    continue

            # Apply expected standard deviation requirement
            if exp_std is not None and std_tolerance is not None:
                delta_std = abs(stddev - exp_std)
                if delta_std > std_tolerance:
                    result.passed = False
                    result.messages += [
                        (
                            f"The observed standard deviation ({stddev}) differs from the expected value by "
                            f"{delta_std} (<={std_tolerance} required)"
                        )
                    ]
                    continue

            # Histogram check successful
            result.messages += [
                f"Histogram of {expression} successfully filled from TTree {key} (contains {h.sum()} events)"
            ]

    return result


def range_check_nd(
    filepath_list,
    expressions,
    limits,
    n_bins,
    blind_ranges,
    tree_pattern,
):
    """N-dimensional range check.

    Produce 2-dimensional histograms of variables taken from a TTree object.

    Args:
        filepath_list (list[file-like]): List of paths to files to analyse
        expressions (dict): Name of the variables (or expressions) to be checked.
        limits (dict): Pre-defined ranges
        n_bins (dict): Number of bins for the histogram
        blind_ranges (dict): Regions to be blinded in the histogram
        tree_pattern (regex): A regular expression for the TTree object to check

    Returns:
        A CheckResult object, which for each tree contains tree_data key/values:
            histograms (list[Hist]): A list of filled histograms of the quantities defined by the expression parameters
            num_entries (float): The total number of entries in the histogram (with blind ranges applied)
    """
    result = CheckResult("range_nd", True, True)

    # Check if the number of variables matches expectations
    length_expr = len(expressions)
    length_limits = len(limits)
    if length_expr != 2 and length_expr != 3:
        result.messages += ["Expected two or three variables."]
        result.passed = False
        result.can_combine = False
        return result
    if length_expr != length_limits:
        result.messages += [
            "For each variable, a corresponding range should be defined."
        ]
        result.passed = False
        result.can_combine = False
        return result

    for filepath in filepath_list:
        trees_opened = []
        with uproot.open(filepath) as f:
            for key, obj in f.items(cycle=False):
                if not isinstance(obj, uproot.TTree):
                    continue
                if not re.fullmatch(tree_pattern, key):
                    continue
                if key in trees_opened:
                    continue
                trees_opened.append(key)

                # First time: initialise the CheckResult
                if key not in result.tree_data:
                    result.tree_data[key] = {}
                    result.tree_data[key]["histograms"] = []
                    for key_i, key_j in combinations(list(expressions.keys()), 2):
                        axis0 = hist.axis.Regular(
                            n_bins[key_i],
                            limits[key_i]["min"],
                            limits[key_i]["max"],
                            name=expressions[key_i],
                        )
                        axis1 = hist.axis.Regular(
                            n_bins[key_j],
                            limits[key_j]["min"],
                            limits[key_j]["max"],
                            name=expressions[key_j],
                        )
                        h = Hist(
                            axis0,
                            axis1,
                            name=f"{key} {expressions[key_i]}/{expressions[key_j]}",
                        )
                        result.tree_data[key]["histograms"] += [h]
                    if length_expr == 3:
                        for key_i, key_j, key_k in combinations(
                            list(expressions.keys()), 3
                        ):
                            axis0 = hist.axis.Regular(
                                n_bins[key_i],
                                limits[key_i]["min"],
                                limits[key_i]["max"],
                                name=expressions[key_i],
                            )
                            axis1 = hist.axis.Regular(
                                n_bins[key_j],
                                limits[key_j]["min"],
                                limits[key_j]["max"],
                                name=expressions[key_j],
                            )
                            axis2 = hist.axis.Regular(
                                n_bins[key_k],
                                limits[key_k]["min"],
                                limits[key_k]["max"],
                                name=expressions[key_k],
                            )
                            h = Hist(
                                axis0,
                                axis1,
                                axis2,
                                name=f"{key} {expressions[key_i]}/{expressions[key_j]}/{expressions[key_k]}",
                            )
                        result.tree_data[key]["histograms"] += [h]

                    result.tree_data[key]["num_entries"] = 0

                values_obj = {}
                list_expressions = list(expressions.values())
                # Check if the branch is present in the TTree or if the expressions are correctly written
                try:
                    values_obj = obj.arrays(list_expressions, library="np")
                except uproot.exceptions.KeyInFileError as e:
                    result.messages += [f"Missing branch in {key!r} with {e!r}"]
                    result.passed = False
                    result.can_combine = False
                    continue

                # Apply blinding
                if blind_ranges is not None:
                    mask_total = []
                    if isinstance(blind_ranges, dict):
                        # Take into account that there could be multiple regions to blind
                        blind_ranges = [blind_ranges]
                    for blind_range in blind_ranges:
                        indexes = []
                        mask = numpy.ones(
                            len(values_obj[list_expressions[0]]), dtype=bool
                        )
                        for ax, range_limits in blind_range.items():
                            lower, upper = range_limits["min"], range_limits["max"]
                            indexes = numpy.where(
                                (
                                    (values_obj[expressions[ax]] < upper)
                                    & (values_obj[expressions[ax]] > lower)
                                )
                            )
                            mask_tmp = numpy.zeros(
                                len(values_obj[list_expressions[0]]), dtype=bool
                            )
                            mask_tmp[indexes] = True
                            mask = numpy.logical_and(mask, mask_tmp)
                        mask_total.append(mask)
                    mask_final = numpy.zeros(
                        len(values_obj[list_expressions[0]]), dtype=bool
                    )
                    for mask in mask_total:
                        mask_final = numpy.logical_or(mask_final, mask)
                    for expr in expressions.values():
                        values_obj[expr] = values_obj[expr][~mask_final]

                # Fill the histograms
                list_keys = list(expressions.keys())
                hist_index = 0
                for key_i, key_j in combinations(list_keys, 2):
                    h = result.tree_data[key]["histograms"][hist_index]
                    h.fill(
                        values_obj[expressions[key_i]], values_obj[expressions[key_j]]
                    )
                    hist_index += 1
                # If more than two variables are given in input, return also 3D histograms
                if length_expr == 3:
                    for key_i, key_j, key_k in combinations(list_keys, 3):
                        h = result.tree_data[key]["histograms"][hist_index]
                        h.fill(
                            values_obj[expressions[key_i]],
                            values_obj[expressions[key_j]],
                            values_obj[expressions[key_k]],
                        )
                        hist_index += 1

                # Add to event counter
                result.tree_data[key]["num_entries"] += values_obj[
                    list_expressions[0]
                ].size

    # If no matches are found the check should be marked as failed, and can return early
    if len(result.tree_data) == 0:
        result.passed = False
        result.can_combine = False
        result.messages += [f"No TTree objects found that match {tree_pattern}"]
        return result

    # Check the completely filled histograms
    if result.passed:
        for key in result.tree_data:
            for h in result.tree_data[key]["histograms"]:

                # Require at least one event
                if h.sum() == 0:
                    result.passed = False
                    result.messages += [f"No events found in range for Tree {key}"]
                    continue
                # Histogram check successful
                if len(h.axes) == 2:
                    message = f"Histogram of {h.axes[0].name}, {h.axes[1].name} successfully filled"
                    message += f" from TTree {key} (contains {h.sum()} events)"
                    result.messages += [message]
                else:
                    message = f"Histogram of {h.axes[0].name}, {h.axes[1].name}, {h.axes[2].name} successfully filled"
                    message += f" from TTree {key} (contains {h.sum()} events)"
                    result.messages += [message]
    return result


def num_entries_per_invpb(
    filepath_list,
    count_per_invpb,
    tree_pattern,
    lumi_pattern,
):
    """Number of entries per inverse picobarn check.

    Check that the matching TTree objects contain a minimum number of entries per unit luminosity (pb-1).

    Args:
        filepath_list (list[file-like]): List of paths to files to analyse
        count_per_invpb (float): The minimum number of entries per unit luminosity required
        tree_pattern (regex): A regular expression for the TTree objects to check
        lumi_pattern (regex): A regular expression for the TTree object containing the luminosity information

    Returns:
        A CheckResult object, which for each tree contains tree_data key/values:
            num_entries (float): The total number of events in the TTree
            lumi_invpb (float): The total luminosity, in inverse picobarns
            num_entries_per_invpb (float): The total number of events divided by the total luminosity
    """
    result = CheckResult("num_entries_per_invpb", True, True)

    lumi = 0
    for filepath in filepath_list:
        trees_opened = []
        with uproot.open(filepath) as f:
            for key, obj in f.items(cycle=False):
                if not isinstance(obj, uproot.TTree):
                    continue

                # If object is the decay TTree
                if re.fullmatch(tree_pattern, key):
                    if key in trees_opened:
                        continue
                    trees_opened.append(key)

                    # First time: initialise the CheckResult
                    if key not in result.tree_data:
                        result.tree_data[key] = {}
                        result.tree_data[key]["num_entries"] = 0
                        result.tree_data[key]["lumi_invpb"] = 0
                        result.tree_data[key]["num_entries_per_invpb"] = None

                    # Add number of entries to counter
                    result.tree_data[key]["num_entries"] += obj.num_entries

                # If object is lumi TTree
                if re.fullmatch(lumi_pattern, key):
                    try:
                        lumi_arr = obj["IntegratedLuminosity"].array(library="np")
                        err_lumi_arr = obj["IntegratedLuminosityErr"].array(
                            library="np"
                        )
                    except uproot.exceptions.KeyInFileError as e:
                        result.messages += [
                            f"Missing luminosity branch in {key!r} with error {e!r}"
                        ]
                        result.passed = False
                        result.can_combine = False
                        break

                    err_lumi_quad_sum = math.sqrt(numpy.sum(err_lumi_arr**2))
                    if err_lumi_quad_sum / numpy.sum(lumi_arr) >= 1:
                        result.passed = False
                        result.can_combine = False
                        result.messages += [
                            "Luminosity information is not reliable: 100% or greater relative uncertainty"
                        ]
                        break
                    # Add to luminosity counter
                    lumi += numpy.sum(lumi_arr)

    # If no matches are found the check should be marked as failed, and can return early
    if len(result.tree_data) == 0:
        result.passed = False
        result.can_combine = False
        result.messages += [f"No TTree objects found that match {tree_pattern}"]
        return result

    if result.passed:
        for key in result.tree_data:
            if lumi == 0:
                result.passed = False
                result.can_combine = False
                result.messages += [
                    "Failed to get luminosity information (total luminosity = 0)"
                ]
                continue

            entries_per_lumi = round(result.tree_data[key]["num_entries"] / lumi, 2)
            result.tree_data[key]["lumi_invpb"] = lumi
            result.tree_data[key]["num_entries_per_invpb"] = entries_per_lumi
            if entries_per_lumi < count_per_invpb:
                result.passed = False
            result.messages += [
                f"Found {entries_per_lumi} entries per unit luminosity (pb-1) in {key} ({count_per_invpb} required)"
            ]

    return result


def range_check_bkg_subtracted(
    filepath_list,
    expression,
    limits,
    expr_for_subtraction,
    mean_sig,
    background_shift,
    background_window,
    signal_window,
    n_bins,
    blind_ranges,
    tree_pattern,
):
    """Range check with background subtraction.

    Check if there is at least one entry in the TTree object with a specific variable falling in a pre-defined range.
    The background-subtracted histogram is then produced as output. Background is subtracted assuming a linear
    distribution. In particular, signal ([m-s, m+s]) and background ([m-b-delta, m-b] U [m+b, m+b+delta]) windows have
    to be defined on a control variable. Then, one histogram is created for events falling in the signal region and
    another histogram is created for events falling in the background region. The subtraction, using the proper scaling
    factor, is finally performed. It is also possible to blind some regions.

    Args:
        filepath_list (list[file-like]): List of paths to files to analyse
        expression (str): Name of the variable (or expression depending on varibales in the TTree) to be checked
        limits (dict): Pre-defined range
        expr_for_subtraction (str): Name of the control variable (or expression depending on varibales in the TTree) to
            be used to perform background subtraction
        mean_sig (float): Expected mean value of expr_for_subtraction variable. The signal window will be centered
            around this value.
        background_shift (float):  Shift, w.r.t the "mean_sig" value, used to define the two background regions.
        background_window (float):  Length of the background windows (of expr_for_subtraction variable).
        signal_window (float): Length of the signal window (of expr_for_subtraction variable) used for background
            subtraction. The window is centered around the value of "mean_sig".
        n_bins (int): Number of bins for the histogram
        blind_ranges (dict): Regions to be blinded in the histogram
        tree_pattern (regex): A regular expression for the TTree object to check

    Returns:
        A CheckResult object, which for each tree contains tree_data key/values:
            histograms: A list of filled 1D histograms,
                Index 0: The control variable used to perform the subtraction
                Index 1: Events in the signal window
                Index 2: Events in the background window
                Index 3: The background-subtracted result
    """
    result = CheckResult("range_bkg_subtracted", True, True)

    for filepath in filepath_list:
        trees_opened = []
        with uproot.open(filepath) as f:
            for key, obj in f.items(cycle=False):
                if not isinstance(obj, uproot.TTree):
                    continue
                if not re.fullmatch(tree_pattern, key):
                    continue
                if key in trees_opened:
                    continue
                trees_opened.append(key)

                # Calculate the min and max values of each of the two background regions.
                # By construction, the two intervals have the same length
                background_range_low = {
                    "min": mean_sig - background_shift - background_window,
                    "max": mean_sig - background_shift,
                }
                background_range_high = {
                    "min": mean_sig + background_shift,
                    "max": mean_sig + background_shift + background_window,
                }
                # Calculate the min and max values of each of the signal region
                signal_range = {
                    "min": mean_sig - signal_window / 2.0,
                    "max": mean_sig + signal_window / 2.0,
                }
                # First time: initialise the CheckResult
                if key not in result.tree_data:
                    result.tree_data[key] = {}

                    # Create the histogram for the control variable used to perform background subtraction
                    axis0 = hist.axis.Regular(
                        n_bins,
                        background_range_low["min"],
                        background_range_high["max"],
                        name=expr_for_subtraction,
                    )
                    result.tree_data[key]["histograms"] = [
                        Hist(
                            axis0,
                            name=f"{key} {expression}",
                            storage=hist.storage.Weight(),
                        )
                    ]
                    # Add ranges to histogram metadata. Signal and background regions can be then highlighted in the final plot.
                    result.tree_data[key]["histograms"][0].metadata = [
                        background_range_low["min"],
                        background_range_low["max"],
                        background_range_high["min"],
                        background_range_high["max"],
                        signal_range["min"],
                        signal_range["max"],
                    ]

                    # Add signal & background region histograms (using same axis)
                    axis1 = hist.axis.Regular(
                        n_bins, limits["min"], limits["max"], name=expression
                    )
                    result.tree_data[key]["histograms"].append(
                        Hist(
                            axis1,
                            name=f"{key} {expression} signal",
                            storage=hist.storage.Weight(),
                        )
                    )
                    result.tree_data[key]["histograms"].append(
                        Hist(
                            axis1,
                            name=f"{key} {expression} background",
                            storage=hist.storage.Weight(),
                        )
                    )
                    result.tree_data[key]["histograms"].append(
                        Hist(
                            axis1,
                            name=f"{key} {expression} signal after bkg subtraction",
                            storage=hist.storage.Weight(),
                        )
                    )

                values_obj = {}
                # Check if the branch is in the Tree or if the expressions are correctly written
                try:
                    values_obj = obj.arrays(
                        [expr_for_subtraction, expression], library="np"
                    )
                except uproot.exceptions.KeyInFileError as e:
                    result.messages += [f"Missing branch in {key!r} with {e!r}"]
                    result.passed = False
                    result.can_combine = False
                    continue

                # Fill control variable histogram
                var_for_bkgsub_array = values_obj[expr_for_subtraction]
                result.tree_data[key]["histograms"][0].fill(var_for_bkgsub_array)

                # Select events in signal region
                cut_string = (
                    "("
                    + expr_for_subtraction
                    + ">"
                    + str(signal_range["min"])
                    + ") & ("
                    + expr_for_subtraction
                    + "<"
                    + str(signal_range["max"])
                    + ")"
                )
                values_sig = obj.arrays([expression], cut_string, library="np")
                test_array_sig = values_sig[expression]
                test_array_sig = test_array_sig[
                    numpy.where(
                        (test_array_sig < limits["max"])
                        & (test_array_sig > limits["min"])
                    )
                ]
                # Select events in background region
                cut_string = (
                    "( ("
                    + expr_for_subtraction
                    + ">"
                    + str(background_range_low["min"])
                    + ") & ("
                    + expr_for_subtraction
                    + "<"
                    + str(background_range_low["max"])
                    + ") ) | ( ("
                    + expr_for_subtraction
                    + ">"
                    + str(background_range_high["min"])
                    + ") & ("
                    + expr_for_subtraction
                    + "<"
                    + str(background_range_high["max"])
                    + ") )"
                )
                values_bkg = obj.arrays([expression], cut_string, library="np")
                test_array_bkg = values_bkg[expression]
                test_array_bkg = test_array_bkg[
                    numpy.where(
                        (test_array_bkg < limits["max"])
                        & (test_array_bkg > limits["min"])
                    )
                ]

                # Apply blinding
                if blind_ranges is not None:
                    if isinstance(blind_ranges, dict):
                        blind_ranges = [blind_ranges]
                        # Take into account that there could be multiple regions to blind
                    for blind_range in blind_ranges:
                        lower, upper = blind_range["min"], blind_range["max"]
                        test_array_sig = test_array_sig[
                            ~((lower < test_array_sig) & (test_array_sig < upper))
                        ]
                        test_array_bkg = test_array_bkg[
                            ~((lower < test_array_bkg) & (test_array_bkg < upper))
                        ]

                # Fill signal & background histograms
                result.tree_data[key]["histograms"][1].fill(test_array_sig)
                result.tree_data[key]["histograms"][2].fill(test_array_bkg)

    # If no matches are found the check should be marked as failed, and can return early
    if len(result.tree_data) == 0:
        result.passed = False
        result.can_combine = False
        result.messages += [f"No TTree objects found that match {tree_pattern}"]
        return result

    if result.passed:
        for key in result.tree_data:

            # Require events in both signal and background histograms
            if (result.tree_data[key]["histograms"][1].view().value.sum() == 0) or (
                result.tree_data[key]["histograms"][2].view().value.sum() == 0
            ):
                result.passed = False
                result.messages += [
                    f"Not enough events for background subtraction found in range for Tree {key}"
                ]
                continue

            # Assume linear background distribution and evaluate fraction of background in the signal region
            alpha = 2.0 * background_window / signal_window

            # Histogram subtraction
            hsub = (
                result.tree_data[key]["histograms"][1]
                + (-1 * alpha) * result.tree_data[key]["histograms"][2]
            )
            result.tree_data[key]["histograms"][3] = hsub

            result.messages += [
                f"Background subtraction performed successfully for Tree {key}"
            ]

    return result


def branches_exist(
    filepath_list,
    branches,
    tree_pattern,
):
    """Branches exist check.

    Check that all matching TTree objects contain a minimum number of entries.

    Args:
        filepath_list: List of paths to files to analyse
        branches: List of branches that will be required to exist in TTree objects
        tree_pattern: A regular expression for the TTree objects to check

    Returns:
        A CheckResult object, which for each tree contains no tree_data key/values (an empty dict)
    """
    result = CheckResult("branches_exist", True, True)
    for filepath in filepath_list:
        trees_opened = []
        with uproot.open(filepath) as f:
            for key, obj in f.items(cycle=False):
                if not isinstance(obj, uproot.TTree):
                    continue
                if not re.fullmatch(tree_pattern, key):
                    continue
                if key in trees_opened:
                    continue
                trees_opened.append(key)

                # First time: initialise the CheckResult
                if key not in result.tree_data:
                    result.tree_data[key] = {}

                # Check that branches exist
                if not set(branches).issubset(obj.keys()):
                    missing_branches = list(set(branches) - set(obj.keys()))
                    result.passed = False
                    result.can_combine = False
                    result.messages += [
                        f"Required branches not found in Tree {key}: {missing_branches}"
                    ]

    # If no matches are found the check should be marked as failed, and can return early
    if len(result.tree_data) == 0:
        result.passed = False
        result.can_combine = False
        result.messages += [f"No TTree objects found that match {tree_pattern}"]
        return result

    if result.passed:
        for key, _data in result.tree_data.items():
            result.messages += [f"All required branches were found in Tree {key}"]

    return result



###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import re
from collections import OrderedDict, defaultdict
from os.path import isfile, join, relpath

import jinja2
import yaml
from strictyaml import Any, Bool, Enum, Float, Int, Map, MapPattern
from strictyaml import Optional
from strictyaml import Optional as Opt
from strictyaml import Regex, Seq, Str, load

from LbAPCommon import config
from LbAPCommon.linting.bk_paths import validate_bk_query

RE_APPLICATION = r"^([A-Za-z]+/)+v\d+r\d+(p\d+)?"
RE_JOB_NAME = r"^[a-zA-Z0-9][a-zA-Z0-9_\-]+$"
RE_OUTPUT_FILE_TYPE = r"^([A-Za-z][A-Za-z0-9_]+\.)+((ROOT|root)|.?(DST|dst))$"
RE_OPTIONS_FN = r"^\$?[a-zA-Z0-9/\.\-\+\=_]+$"
RE_INFORM = r"^(?:[a-zA-Z]{3,}|[^@\s]+@[^@\s]+\.[^@\s]+)$"

RE_ROOT_IN_TES = r"^\/.+$"
RE_DDDB_TAG = r"^.{1,50}$"
RE_CONDDB_TAG = r"^.{1,50}$"

RE_COMMENT = r"(.{1,5000})"
DQ_FLAGS_SCHEMA = Seq(Enum(["BAD", "OK", "EXPRESS_OK", "UNCHECKED"]))

LEGACY_OPTIONS = {
    Optional("command"): Seq(Str()),
    "files": Seq(Regex(RE_OPTIONS_FN)),
}
LBEXEC_OPTIONS = {
    "entrypoint": Regex(r".+:.+"),
    Optional("extra_options"): MapPattern(Str(), Any()),
    Optional("extra_args"): Seq(Str()),
}

BASE_JOB_SCHEMA = {
    "application": Regex(RE_APPLICATION),
    "input": MapPattern(Str(), Any()),
    "output": Regex(RE_OUTPUT_FILE_TYPE) | Seq(Regex(RE_OUTPUT_FILE_TYPE)),
    "options": Regex(RE_OPTIONS_FN)
    | Seq(Regex(RE_OPTIONS_FN))
    | MapPattern(Str(), Any()),
    "wg": Enum(config.known_working_groups),
    "inform": Regex(RE_INFORM) | Seq(Regex(RE_INFORM)),
    # Automatic configuration
    "automatically_configure": Bool(),
    "turbo": Bool(),
    Optional("root_in_tes"): Regex(RE_ROOT_IN_TES),
    Optional("simulation"): Bool(),
    Optional("luminosity"): Bool(),
    Optional("data_type"): Enum(config.known_data_types),
    Optional("input_type"): Enum(config.known_input_types),
    Optional("dddb_tag"): Regex(RE_DDDB_TAG),
    Optional("conddb_tag"): Regex(RE_CONDDB_TAG),
    Optional("checks"): Seq(Str()),  # TODO: replace this with a regex
    Optional("extra_checks"): Seq(Str()),  # TODO: replace this with a regex
    # Production submission metadata
    Optional("comment"): Regex(RE_COMMENT),
    "priority": Enum(config.allowed_priorities),
    "completion_percentage": Float(),
}
INPUT_SCHEMAS = {
    "bk_query": Map(
        {"bk_query": Str(), Opt("n_test_lfns"): Int(), Opt("dq_flags"): DQ_FLAGS_SCHEMA}
    ),
    "job_name": Map({"job_name": Str(), Opt("filetype"): Regex(RE_OUTPUT_FILE_TYPE)}),
    "transform_ids": Map(
        {
            "transform_ids": Seq(Int()),
            "filetype": Regex(RE_OUTPUT_FILE_TYPE),
            Opt("n_test_lfns"): Int(),
            Opt("dq_flags"): DQ_FLAGS_SCHEMA,
        }
    ),
}
DEFAULT_JOB_VALUES = {
    "automatically_configure": False,
    "turbo": False,
    "completion_percentage": 100,
    "priority": "1b",
}

CHECK_TYPE_SCHEMAS = {
    "range": {
        "expression": Str(),  # TODO: replace this with a regex
        "limits": Map({"min": Float(), "max": Float()}),
        Optional("n_bins"): Int(),
        Optional("blind_ranges"): Map({"min": Float(), "max": Float()})
        | Seq(Map({"min": Float(), "max": Float()})),
        Optional("exp_mean"): Float(),
        Optional("exp_std"): Float(),
        Optional("mean_tolerance"): Float(),
        Optional("std_tolerance"): Float(),
    },
    "range_nd": {
        "expressions": Map(
            {  # TODO: replace Str() with a regex
                "x": Str(),
                "y": Str(),
                Optional("z"): Str(),
            }
        ),
        "limits": Map(
            {
                "x": Map({"min": Float(), "max": Float()}),
                "y": Map({"min": Float(), "max": Float()}),
                Optional("z"): Map({"min": Float(), "max": Float()}),
            }
        ),
        Optional("n_bins"): Map(
            {
                "x": Int(),
                "y": Int(),
                Optional("z"): Int(),
            }
        ),
        Optional("blind_ranges"): Seq(
            Map(
                {
                    "x": Map({"min": Float(), "max": Float()}),
                    "y": Map({"min": Float(), "max": Float()}),
                    Optional("z"): Map({"min": Float(), "max": Float()}),
                }
            )
        ),
    },
    "num_entries": {
        "count": Int(),
    },
    "num_entries_per_invpb": {
        "count_per_invpb": Float(),
        Optional("lumi_pattern"): Str(),
    },
    "range_bkg_subtracted": {
        "expression": Str(),
        "limits": Map({"min": Float(), "max": Float()}),
        "expr_for_subtraction": Str(),
        "mean_sig": Float(),
        "background_shift": Float(),
        "background_window": Float(),
        "signal_window": Float(),
        Optional("n_bins"): Int(),
        Optional("blind_ranges"): Map({"min": Float(), "max": Float()})
        | Seq(Map({"min": Float(), "max": Float()})),
    },
    "branches_exist": {
        "branches": Seq(Str()),
    },
}
BASE_CHECK_SCHEMA = {
    "type": Enum(list(CHECK_TYPE_SCHEMAS)),
    Optional("tree_pattern"): Str(),
}
BASE_CHECK_DEFAULT_VALUES = {
    "tree_pattern": r"(.*/DecayTree)|(.*/MCDecayTree)",
}
CHECK_TYPE_DEFAULT_VALUES = {
    "num_entries": {},
    "range": {
        "n_bins": 50,
    },
    "range_nd": {
        "n_bins": {
            "x": 50,
            "y": 50,
            "z": 50,
        },
    },
    "num_entries_per_invpb": {
        "lumi_pattern": r"(.*/LumiTuple)",
    },
    "range_bkg_subtracted": {
        "n_bins": 50,
    },
    "branches_exist": {},
}


def _ordered_dict_to_dict(a):
    if isinstance(a, (OrderedDict, dict)):
        return {k: _ordered_dict_to_dict(v) for k, v in a.items()}
    elif isinstance(a, (list, tuple)):
        return [_ordered_dict_to_dict(v) for v in a]
    else:
        return a


def render_yaml(raw_yaml):
    """Render a "raw" YAML jinja template.

    Accepts LbAP yaml configuration jinja template and renders it into a full YAML configuration.

    Args:
        raw_yaml (str): YAML jinja-template string

    Raises:
        ValueError: raised if jinja2 couldn't render the raw_yaml string.

    Returns:
        str: a jinja-rendered YAML string.
    """
    try:
        rendered_yaml = jinja2.Template(
            raw_yaml, undefined=jinja2.StrictUndefined
        ).render()
    except jinja2.TemplateError as e:
        raise ValueError(
            "Failed to render with jinja2 on line %s: %s"
            % (getattr(e, "lineno", "unknown"), e)
        ) from e
    return rendered_yaml


def _validate_proc_pass_map(job_names, proc_pass_map):
    """Build processing paths and validate them from a processing pass map.

    Given a list of step job names (in correct order), and the processing pass map,
    build the processing path for each step and verify the length is below 100.

    Args:
        job_names (list[str]): a list containing step job names.
        proc_pass_map (dict): A dictionary mapping job names to processing pass

    Raises:
        ValueError: raised if the processing path length is over 100 characters
    """
    for i, job_name in enumerate(job_names):
        proc_passes = map(proc_pass_map.get, job_names[:i] + [job_name])
        pro_path = "/".join(proc_passes)
        if len(pro_path) >= 100:
            proc_pass = proc_pass_map[job_name]
            step_jobs_list = "  - " + "\n  - ".join(job_names)
            raise ValueError(
                f"The expected processing path length for the job '{job_name}' is too long.\n"
                "DIRAC requires this to be less than 100 characters.\n\n"
                f"'Step' jobs:\n{step_jobs_list}\n"
                f"Job name: {job_name}\n"
                f"Processing pass for this step: {proc_pass}\n"
                f"Processing path for this step ({len(pro_path)} chars): {pro_path}\n\n"
                "To recover from this issue, consider:"
                "  - Removing redundant information from your job name.\n"
                "  - Shortening your job names.\n"
                "  - If the offending job depends on output from other jobs, ensure that they have a common prefix.\n"
            )


def create_proc_pass_map(job_names, version, default_proc_pass="default"):
    """Create a job name to processing pass map.

    Given a list of step job names and the production version, produce a
    job_name --> processing pass mapping.

    Args:
        job_names (list): step job names
        version (str): LbAPproduction version
        default_proc_pass (str, optional): the default processing pass. Defaults to "default".

    Returns:
        dict: a step job name to processing pass map
    """
    proc_pass_prefix = f"AnaProd-{version}-"
    proc_pass_map = {}

    # dummy_version = "v0r0p00000000"

    def clean_proc_pass(i, original_job_name):
        # attempt to remove redundant information from the job name
        job_name = re.sub(
            r"([0-9]{8})|(MagUp|MagDown|MU|MD)|((^|[^0*9])201[125678]($|[^0*9]))",
            "",
            original_job_name,
        )
        # Remove repeated separator chatacters
        job_name = re.sub(r"([-_])[-_]+", r"\1", job_name).strip("_-")
        if i == 0:
            return f"{proc_pass_prefix}{job_name}"

        proc_pass = job_name
        for previous_job_name in job_names[:i]:
            size = 0
            previous_proc_pass = proc_pass_map[previous_job_name]
            # Remove the prefix if this is the first job
            if previous_proc_pass.startswith(proc_pass_prefix):
                previous_proc_pass = previous_proc_pass[len(proc_pass_prefix) :]
            # Look for a common prefix and remove it
            for last, this in zip(previous_proc_pass, proc_pass):
                if last != this:
                    break
                size += 1
            proc_pass = proc_pass[size:].strip("_-+")
            # If the processing pass has been entirely stripped use a default
            if not proc_pass:
                proc_pass = default_proc_pass

        return proc_pass

    for i, job_name in enumerate(job_names):
        proc_pass_map[job_name] = clean_proc_pass(i, job_name)

    _validate_proc_pass_map(job_names, proc_pass_map)

    return proc_pass_map


def is_simulation_job(prod_data: dict, job_name: str):
    """Determine if a job is using MC input or not.

    Args:
        prod_data (dict): Entire production information from yaml parsing
        job_name (str): Name of the job to determine if it's using MC input or not

    Raises:
        NotImplementedError: No bookkeeping location or job name provided.

    Returns:
        bool: True if the job is using MC input, False if it is not
    """
    job_dict = prod_data[job_name]
    if "simulation" not in job_dict:
        if "bk_query" in job_dict["input"]:
            if "/mc/" in job_dict["input"]["bk_query"].lower():
                return True
            else:
                return False
        elif "job_name" in job_dict["input"]:
            dependent_job = prod_data[job_name]["input"]["job_name"]
            return is_simulation_job(prod_data, dependent_job)
        else:
            raise NotImplementedError(
                "Input requires either a bookkeeping location or a previous job name"
            )


def parse_yaml(rendered_yaml):
    """Parse rendered YAML text.

    Args:
        rendered_yaml (str): The rendered YAML jinja template.

    Raises:
        ValueError: raised if errors occurred during parsing.

    Returns:
        tuple: a tuple of the parsed configuration data2 (dict) and checks data (dict).
    """
    data1 = load(
        rendered_yaml, schema=MapPattern(Regex(RE_JOB_NAME), Any(), minimum_keys=1)
    )

    data_checks = {}
    if "checks" in data1:
        # apply the appropriate schema to each different type of check
        for _check_name, check_data in data1["checks"].items():
            check_type = str(check_data["type"])
            check_schema = {
                **BASE_CHECK_SCHEMA,
                **CHECK_TYPE_SCHEMAS[check_type],
            }
            check_defaults = {
                **BASE_CHECK_DEFAULT_VALUES,
                **CHECK_TYPE_DEFAULT_VALUES[check_type],
            }

            # apply default values common to all check types
            for key, value in check_schema.copy().items():
                if isinstance(key, Optional):
                    key = key.key
                    if key not in check_data and key in check_defaults:
                        key = Optional(key, default=check_defaults.get(key))
                        check_schema[key] = value

            check_data.revalidate(Map(check_schema))

        # if checks pass validation, store elsewhere & delete from main data
        # so that normal jobs aren't impacted
        data_checks = data1.data["checks"]

        # special case: range_nd defaults with variable number of axes
        # if only 2D, remove the defaults for the 3rd axis
        for chk, chk_data in data_checks.items():
            if chk_data.get("type") == "range_nd":
                if (
                    chk_data.get("n_bins")
                    == CHECK_TYPE_DEFAULT_VALUES["range_nd"]["n_bins"]
                ):
                    if len(chk_data.get("expressions")) == 2:
                        del data_checks[chk]["n_bins"]["z"]

        del data1["checks"]

    if "defaults" in data1:
        defaults_schema = {}
        for key, value in BASE_JOB_SCHEMA.items():
            if isinstance(key, Optional):
                key = key.key
            key = Optional(key, default=DEFAULT_JOB_VALUES.get(key))
            defaults_schema[key] = value

        data1["defaults"].revalidate(Map(defaults_schema))
        defaults = data1.data["defaults"]
        # Remove the defaults data from the snippet
        del data1["defaults"]
    else:
        defaults = DEFAULT_JOB_VALUES.copy()

    job_names = list(data1.data.keys())
    if len(set(n.lower() for n in job_names)) != len(job_names):
        raise ValueError(
            "Found multiple jobs with the same name but different capitalisation"
        )

    job_name_schema = Regex(r"(" + r"|".join(map(re.escape, job_names)) + r")")

    # StrictYAML has non-linear complexity when parsing many keys
    # Avoid extremely slow parsing by doing each key individually
    data2 = {}
    for k, v in data1.items():
        k = k.data
        v = _ordered_dict_to_dict(v.data)

        production_schema = {}
        if "comment" in v:
            raise ValueError(
                "comment is only allowed to be set in the defaults of the production!"
            )
        for key, value in BASE_JOB_SCHEMA.items():
            if isinstance(key, Optional):
                key = key.key
                production_schema[Optional(key, default=defaults.get(key))] = value
            elif key in defaults:
                production_schema[Optional(key, default=defaults[key])] = value
            else:
                production_schema[key] = value

        data = load(
            yaml.safe_dump({k: v}),
            MapPattern(job_name_schema, Map(production_schema), minimum_keys=1),
        )
        for input_key, input_schema in INPUT_SCHEMAS.items():
            if input_key in data.data[k]["input"]:
                data[k]["input"].revalidate(input_schema)
                break
        else:
            raise ValueError(
                (
                    "Failed to find a valid schema for %s's input. "
                    "Allowed values are: %s"
                )
                % (k, set(INPUT_SCHEMAS))
            )
        if isinstance(data.data[k]["options"], dict):
            if "files" in data.data[k]["options"]:
                data[k]["options"].revalidate(Map(LEGACY_OPTIONS))
            else:
                data[k]["options"].revalidate(Map(LBEXEC_OPTIONS))

        # move contents of extra_checks to checks
        data_dict = data.data
        if "checks" not in data_dict[k]:
            data_dict[k]["checks"] = []
        if "extra_checks" in data_dict[k]:
            data_dict[k]["checks"] += data_dict[k]["extra_checks"]
            del data_dict[k]["extra_checks"]

        data2.update(data_dict)

    return data2, data_checks


def _normalise_filetype(prod_name, job_name, filetype):
    filetype = filetype.upper()

    errors = []
    if len(filetype) >= 50:
        errors += ["The filetype is excessively long"]
    if re.findall(r"[0-9]{8}", filetype, re.IGNORECASE):
        errors += ["It appears the event type is included"]
    if re.findall(r"Mag(Up|Down)", filetype, re.IGNORECASE):
        errors += ["It appears the magnet polarity is included"]
    if re.findall(r"(^|[^0*9])201[125678]($|[^0*9])", filetype, re.IGNORECASE):
        errors += ["It appears the data taking year is included"]

    if errors:
        _errors = "\n  * ".join(errors)
        raise ValueError(
            f"Output filetype {filetype} for {prod_name}/{job_name} is invalid "
            f"as it appears to contain redundant information.\n\n"
            f"  * {_errors}"
        )
    return filetype


def _check_name_magnet_polarity(bk_query, job_name):
    match = re.search(r"-mag(up|down)[-/]", bk_query)
    if not match:
        return [f"Failed to find magnet polarity in {bk_query}"]
    good_pol = match.groups()[0]
    bad_pol = {"down": "up", "up": "down"}[good_pol]
    if f"mag{bad_pol}" in job_name:
        raise ValueError(
            f"Found 'mag{bad_pol}' in job name {job_name!r} with"
            f"'mag{good_pol}' input ({bk_query!r}). "
            "Has the wrong magnet polarity been used?"
        )
    match = re.search(r"([^a-z0-9]|\b)m(u|d)([^a-z0-9]|\b)", job_name)
    if match and match.groups()[1] == bad_pol[0]:
        raise ValueError(
            f"Found 'm{bad_pol[0]}' in job name {job_name!r} with"
            f" 'mag{good_pol}' input ({bk_query!r}). "
            "Has the wrong magnet polarity been used?"
        )
    return []


def _check_job_chain_name_polarity(job_name: str, input_job_name: str) -> list:
    mags = ["mag", "m"]
    down = ["down", "d"]
    up = ["up", "u"]
    for i, pattern in enumerate(
        [
            r"([^a-z0-9]|\b)mag(up|down)([^a-z0-9]|\b)",
            r"([^a-z0-9]|\b)m(u|d)([^a-z0-9]|\b)",
        ]
    ):
        matches = [
            re.search(pattern, name, re.IGNORECASE)
            for name in [job_name, input_job_name]
        ]
        if matches[1]:
            groups = [group for group in matches[1].groups() if group != "_"]
            good_pol = groups[0].lower()
            bad_pol = {down[i]: up[i], up[i]: down[i]}[good_pol]
            if f"{mags[i]}{bad_pol}" in job_name.lower():
                raise ValueError(
                    f"Found '{mags[i]}{bad_pol}' in job name {job_name!r} with"
                    f" '{mags[i]}{good_pol}' input ({input_job_name!r}). "
                    "Has the wrong magnet polarity been used?"
                )
            break
    else:
        return [
            f"Failed to find magnet polarity in {input_job_name}, "
            "cannot determine if polarities are matched throughout the job chain"
        ]

    return []


def validate_yaml(jobs_data, checks_data, repo_root, prod_name):
    """Validate YAML configuration.

    Args:
        jobs_data (dict): Parsed job configuration.
        checks_data (dict): Parsed checks configuration.
        repo_root (str): Repository location.
        prod_name (str): Production name.

    Raises:
        ValueError: Raised if there are showstopper issues in the parsed job configuration.

    Returns:
        list[str]: warnings (not showstoppers) raised during validation of the configuration.
    """
    bk_query_to_job = defaultdict(set)

    warnings = []

    # Ensure all values that can be either a list or a string are lists of strings
    for job_name, job_data in jobs_data.items():
        try:
            _validate_job_data(repo_root, prod_name, job_name, job_data, checks_data)
        except Exception as e:
            raise ValueError(f"Failed to validate {job_name!r} with error {e!r}") from e

        if "bk_query" in job_data["input"]:
            if job_data["input"]["bk_query"].lower() in bk_query_to_job:
                warnings.append(
                    f'Duplicate bk_query found (this is a case insensitive check): {job_data["input"]["bk_query"].lower()}'
                )
            bk_query_to_job[job_data["input"]["bk_query"].lower()].add(job_name.lower())

        elif "job_name" in job_data["input"]:
            # Ensure that the job name and output filetype are unambiguously defined
            if job_data["input"]["job_name"] not in jobs_data:
                raise ValueError(
                    f"Unrecognised job name in input: {job_data['input']['job_name']}"
                )
            warnings.extend(
                _check_job_chain_name_polarity(job_name, job_data["input"]["job_name"])
            )
            input_job_data = jobs_data[job_data["input"]["job_name"]]
            input_filetype = job_data["input"].get("filetype", "").upper()
            if len(input_job_data["output"]) == 1:
                if input_filetype not in [""] + input_job_data["output"]:
                    raise ValueError(
                        f"Unrecognised {input_filetype=} for {job_name=} input, "
                        f"expected one of: {input_job_data['output']}"
                    )
            elif input_filetype == "":
                raise ValueError(
                    f"{job_name} gets its input from a job with multiple outputs. "
                    "The 'filetype' key must be specified in the 'input' section."
                )
            elif input_filetype.upper() not in input_job_data["output"]:
                raise ValueError(
                    f"Unrecognised {input_filetype=} for {job_name=} input, "
                    f"expected one of: {input_job_data['output']}"
                )

    for query, job_set in bk_query_to_job.items():
        for job_name in job_set:
            warnings.extend(_check_name_magnet_polarity(query, job_name))

    polarity_len_swap = {"magdown": 5, "magup": 7}
    polarity_swap = {"magdown": "magup", "magup": "magdown"}

    # check that both polarities are always used for a given bk_query
    for query in bk_query_to_job:
        index = None
        query_polarity = None
        both_polarities = False
        for polarity in {"magdown", "magup"}:
            if polarity in query:
                index = query.find(polarity)
                query_polarity = polarity
        if query_polarity:
            for compare_query in bk_query_to_job:
                if compare_query != query:
                    if polarity_swap[query_polarity] in compare_query:
                        if (
                            compare_query[:index]
                            + query_polarity
                            + compare_query[
                                (index + polarity_len_swap[query_polarity]) :
                            ]
                            == query
                        ):
                            both_polarities = True
                            if len(bk_query_to_job[query]) != len(
                                bk_query_to_job[compare_query]
                            ):
                                warnings.append(
                                    f"The number of jobs requesting {query} does not"
                                    " match the number of jobs requesting its opposite"
                                    f" polarity counterpart {compare_query}."
                                )
            if not both_polarities:
                warnings.append(
                    f"{query} has been requested as input for {len(bk_query_to_job[query])} job(s)"
                    " but its opposite polarity counterpart has not been requested for any jobs."
                    " Are you sure you do not want the other polarity?"
                )

    # Validate checks
    try:
        _validate_checks_data(checks_data, jobs_data)
    except Exception as e:
        raise ValueError(f"Failed to validate checks with error {e!r}") from e

    return warnings


def _validate_job_data(repo_root, prod_name, job_name, job_data, checks_data):
    # Normalise list/str fields to always be lists
    for prop in ["output", "options", "inform", "checks", "extra_checks"]:
        if not isinstance(job_data.get(prop, []), (list, dict)):
            job_data[prop] = [job_data[prop]]

    # Validate the input data
    if "bk_query" in job_data["input"]:
        validate_bk_query(job_data["input"]["bk_query"])

    # Validate the output filetype
    job_data["output"] = [
        _normalise_filetype(prod_name, job_name, s) for s in job_data["output"]
    ]

    # Normalise the options filenames if we're using a non-PyConf application
    if isinstance(job_data["options"], list):
        job_data["options"] = {"files": job_data["options"]}
    if "files" in job_data["options"]:
        normalised_options = []
        for fn in job_data["options"]["files"]:
            if fn.startswith("$"):
                normalised_options.append(fn)
                continue

            fn_normed = relpath(join(repo_root, fn), start=repo_root)
            if fn_normed.startswith("../"):
                raise ValueError(f"{fn} not found inside {repo_root}")
            if not isfile(join(repo_root, prod_name, fn_normed)):
                raise FileNotFoundError(
                    f"Production {job_name!r} has a missing options file: "
                    f"{join(prod_name, fn_normed)!r}",
                )
            normalised_options.append(
                join("$ANALYSIS_PRODUCTIONS_BASE", prod_name, fn_normed)
            )
        job_data["options"]["files"] = normalised_options

    # Validate the check names
    # All checks listed for jobs should match a check defined in checks_data
    if "checks" in job_data:
        for ck in job_data["checks"]:
            if ck not in list(checks_data.keys()):
                raise ValueError(f"Check {ck} not found in list of defined checks")

    # Validate the completion percentage
    if not (10 <= job_data["completion_percentage"] <= 100):
        raise ValueError(
            f"Validation failed for job '{job_name}', completion_percentage "
            f"was set to '{job_data['completion_percentage']}', allowed "
            "values are in the interval [10, 100]."
        )


def _validate_checks_data(checks_data, jobs_data):
    # All checks defined in checks_data should be used by at least 1 job
    checks_used = {ck: False for ck in checks_data}
    for _job_name, job_data in jobs_data.items():
        if "checks" in job_data:
            for ck in job_data["checks"]:
                if ck in checks_used:
                    checks_used[ck] = True
    for ck, ck_used in checks_used.items():
        if not ck_used:
            raise ValueError(f"Check {ck} is defined but not used")

    # Requirements for numbers of bins in checks that create histograms
    # Must have 1<n<=100  bins on each axis
    n_bins_maximum = 100
    for _cname, cdata in checks_data.items():
        if cdata.get("type") == "range" or cdata.get("type") == "range_bkg_subtracted":
            if cdata.get("n_bins") <= 1 or cdata.get("n_bins") > n_bins_maximum:
                raise ValueError(
                    f"Number of bins for a histogram must be in the interval [2, {n_bins_maximum}]"
                )
        elif cdata.get("type") == "range_nd":
            for _axis, n_bins in cdata.get("n_bins").items():
                if n_bins <= 1 or n_bins > n_bins_maximum:
                    raise ValueError(
                        f"Number of bins for a histogram must be in the interval [2, {n_bins_maximum}] on each axis"
                    )

    # For mean/stddev in range checks: can't specify tolerance without a central value (& vice versa)
    for _cname, cdata in checks_data.items():
        if cdata.get("type") == "range":
            req_pair_tuples = [
                ("exp_mean", "mean_tolerance"),
                ("exp_std", "std_tolerance"),
            ]
            for pair in req_pair_tuples + [tup[::-1] for tup in req_pair_tuples]:
                if (cdata.get(pair[0]) is not None) and (cdata.get(pair[1])) is None:
                    raise ValueError(
                        f"When using '{pair[0]}', must also specify '{pair[1]}'"
                    )



###############################################################################
# (c) Copyright 2020-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

__all__ = [
    "count_log_messages",
    "parse_log",
]

import re
from collections import defaultdict

import pkg_resources
import yaml

MESSAGE_RE = re.compile(
    r"(# )?(?:(\d{4}-\d{2}-\d{2} \d{2}:\d{2}:\d{2} \w+) )?([^\s]+)\s*"
    r"(VERBOSE|DEBUG|INFO|WARNING|ERROR|FATAL|ALWAYS|SUCCESS) +([^\n]+)"
)


known_messages = yaml.safe_load(
    pkg_resources.resource_string(
        "LbAPCommon", "validators/logs/data/known_messages.yaml"
    ).decode()
)


def count_log_messages(log_text):
    """Parse the log text for errors and advice.

    Args:
        log_text (str): Contents of the log file.

    Returns:
        tuple[int]: Tuple containing the number of WARNING/ERROR/FATAL level messages in
        the job log
    """
    log_messages = _split_by_level(log_text)

    n_warning = len(log_messages["WARNING"])
    n_error = len(log_messages["ERROR"])
    n_fatal = len(log_messages["FATAL"])

    return (n_warning, n_error, n_fatal)


def explain_log(log_text):
    """Splits a Gaudi job log by message level.

    Args:
        log_text (str): Contents of the log file.

    Raises:
        NotImplementedError: when a check is not implemented

    Returns:
        tuple[dict[list[int]]]:
            A tuple of three dictionaries corresponding to (explanations,
            suggestion, error) where each is a dictionary of
            messages to a list of log file line numbers.
    """
    log_messages = _split_by_level(log_text)

    explanations = defaultdict(list)
    suggestions = defaultdict(list)
    errors = defaultdict(list)

    for check in known_messages["full_checks"]:
        if re.search(check["regex"], log_text):
            if check.get("fatal", False):
                errors[check["details"]] = None
            else:
                suggestions[check["details"]] = None

    for check in known_messages["line_checks"]:
        if "regex" in check:
            check_func = re.compile(r"^" + check["regex"] + r"$").match
        elif "message" in check:
            check_func = lambda s: check["message"] == s  # NOQA
        else:
            raise NotImplementedError(check)

        for line_no, _, _, _, message in log_messages[check["level"]]:
            if check_func(message):
                if check.get("fatal", False):
                    errors[check["details"]].append(line_no)
                elif not check.get("ignore", False):
                    suggestions[check["details"]].append(line_no)
                else:
                    explanations[check["details"]].append(line_no)

    return list(explanations.items()), list(suggestions.items()), list(errors.items())


def _split_by_level(log_text):
    """Splits a Gaudi job log by message level.

    Args:
        log_text (str): Contents of the log file

    Returns:
        dict[list[tuple]]: Dictionary split by log level. Each element in a list of matched log
        lines of the form (line_no, prefix, timestamp, service, message).
    """
    log_messages = defaultdict(list)

    for i, line in enumerate(log_text.split("\n"), start=1):
        match = MESSAGE_RE.match(line)
        if match:
            prefix, timestamp, service, level, message = match.groups()
            log_messages[level].append((i, prefix, timestamp, service, message))
    return log_messages



###############################################################################
# (c) Copyright 2020-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

__all__ = ["parse_bookkeeping_xml"]

from collections import namedtuple
from datetime import timedelta
from os.path import basename

from lxml import etree

InputFile = namedtuple("InputFile", ["path", "size", "dataset_size"])
OutputFile = namedtuple("OutputFile", ["path", "size", "dataset_size"])


def parse_bookkeeping_xml(fp, expected_input_files):
    """Extract information from the bookkeeping XML.

    Parameters
    ----------
    fp : file-like
        File pointer of the XML summary
    expected_input_files : :obj:`list`
        List of object describing the expected input files

    Returns
    -------
    passed : :obj:`bool`
        True to the XML summary reports the job as passed
    """
    xml_bk = etree.parse(fp, base_url="LbAnalysisProductions/data/xml")
    events_processed = int(
        xml_bk.find('.//TypedParameter[@Name="NumberOfEvents"]').attrib["Value"]
    )
    events_requested = int(
        xml_bk.find('.//TypedParameter[@Name="StatisticsRequested"]').attrib["Value"]
    )
    run_time = timedelta(
        seconds=float(xml_bk.find('.//TypedParameter[@Name="CPUTIME"]').attrib["Value"])
    )
    cpu_norm = float(
        xml_bk.find('.//TypedParameter[@Name="WNCPUHS06"]').attrib["Value"]
    )
    input_files = _find_input_files(xml_bk, expected_input_files)
    output_files = _find_output_files(xml_bk, input_files)
    return (
        events_processed,
        events_requested,
        run_time,
        cpu_norm,
        input_files,
        output_files,
    )


def _find_input_files(xml_bk, expected_input_files):
    input_files = []
    for e in xml_bk.findall(".//InputFile"):
        input_file = expected_input_files(e.attrib["Name"])
        # fn = e.attrib["Name"]
        # for input_file in expected_input_files:
        #     if input_file.path == basename(fn):
        #         break
        # else:
        #     raise NotImplementedError(
        #         "Failed to find input file %s in %r" % (fn, expected_input_files)
        #     )
        input_files += [
            InputFile(
                path=input_file.path,
                size=input_file.size,
                dataset_size=input_file.dataset_size,
            )
        ]
    return input_files


def _find_output_files(xml_bk, input_files):
    output_files = []
    for f in xml_bk.findall(".//OutputFile"):
        if f.attrib["TypeName"] == "LOG":
            continue
        fn = basename(f.attrib["Name"])
        size = float(f.find('.//Parameter[@Name="FileSize"]').attrib["Value"])
        if len(set(f.dataset_size for f in input_files)) != 1:
            raise NotImplementedError(
                "This estimation hasn't been implemented for"
                "jobs with multiple input datasets"
            )
        dataset_size = int(
            size * input_files[0].dataset_size / sum(f.size for f in input_files)
        )
        output_files.append(OutputFile(path=fn, size=size, dataset_size=dataset_size))
    return output_files



###############################################################################
# (c) Copyright 2020-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

__all__ = ["validate_bk_query"]

import os


def validate_bk_query(bk_query):
    """Perform a cursory check that the BK path is sane.

    Args:
        bk_query (str): BKQuery string.

    Raises:
        ValueError: The BKQuery is not valid.

    Returns:
        bool: True if the BKQuery is valid.
    """
    # Make a dict of the query parts
    mydict = _buildBKQuery(bk_query)

    # Check the EventType is there, is an integer and has 8 digits
    eventtype = mydict.get("EventType")
    if isinstance(eventtype, int) and len(str(eventtype)) == 8:
        return True

    message = "The provided BK query is not valid!"
    # The most common mistake is to have the EventType in the ConditionDescription,
    # with the ConditionDescription ending up in the ProcessingPass.
    if len(str(mydict.get("ConditionDescription", ""))) == 8:
        message += "\n\nFound EventType in ConditionDescription - "
        message += "the EventType must go at the end of the query!"

        mydict["EventType"] = mydict["ConditionDescription"]
        new_cond = mydict["ProcessingPass"].split("/")[1]
        mydict["ConditionDescription"] = new_cond
        mydict["ProcessingPass"] = mydict["ProcessingPass"].replace("/" + new_cond, "")
        new_ft = mydict["ProcessingPass"].split("/")[-1]
        mydict["FileType"] = new_ft
        mydict["ProcessingPass"] = mydict["ProcessingPass"].replace(new_ft, "")
        # Now assemble the query
        retPath = "/"
        for v in mydict.values():
            retPath = os.path.join(retPath, *v.split("/"))
        message += f"\n\nSuggested BK path: {retPath}"
    raise ValueError(message)


def _buildBKQuery(bkPath=""):
    """Builds a dictionary from a BK path. Taken from LHCbDirac BKClient.

    Args:
        bkPath (str, optional): Bookkeeping path. Defaults to "".

    Raises:
        ValueError: BK path is invalid.

    Returns:
        dict: Result of the BKQuery.
    """
    bkQueryDict = {}

    if bkPath:
        bkFields = (
            "ConfigName",
            "ConfigVersion",
            "ConditionDescription",
            "ProcessingPass",
            "EventType",
            "FileType",
        )
        url = bkPath.split(":", 1)
        if len(url) == 1:
            bkPath = url[0]
        else:
            if url[0] == "evt":
                bkFields = (
                    "ConfigName",
                    "ConfigVersion",
                    "EventType",
                    "ConditionDescription",
                    "ProcessingPass",
                    "FileType",
                )
            elif url[0] == "pp":
                bkFields = ("ProcessingPass", "EventType", "FileType")
            elif url[0] == "prod":
                bkFields = ("Production", "ProcessingPass", "EventType", "FileType")
            elif url[0] == "runs":
                bkFields = ("Runs", "ProcessingPass", "EventType", "FileType")
            elif url[0] not in ("sim", "daq", "cond"):
                raise ValueError("Invalid BK path:%s" % bkPath)
            bkPath = url[1]
        if bkPath[0] != "/":
            bkPath = "/" + bkPath
        if bkPath[0:2] == "//":
            bkPath = bkPath[1:]
        bkPath = bkPath.replace("RealData", "Real Data")
        i = 0
        processingPass = "/"
        defaultPP = False
        bk = bkPath.split("/")[1:] + len(bkFields) * [""]
        for bpath in bk:
            if bkFields[i] == "ProcessingPass":
                if (
                    bpath != ""
                    and bpath.upper() != "ALL"
                    and not bpath.split(",")[0].split(" ")[0].isdigit()
                ):
                    processingPass = os.path.join(processingPass, bpath)
                    continue
                # Set the PP
                if processingPass != "/":
                    bkQueryDict["ProcessingPass"] = processingPass
                else:
                    defaultPP = True
                i += 1
            if bkFields[i] == "EventType" and bpath:
                eventTypeList = []
                if bpath.upper() == "ALL":
                    bpath = "ALL"
                else:
                    for et in bpath.split(","):
                        try:
                            eventType = int(et.split(" ")[0])
                            eventTypeList.append(eventType)
                        except ValueError:
                            pass
                    if len(eventTypeList) == 1:
                        eventTypeList = eventTypeList[0]
                    bpath = eventTypeList
            # Set the BK dictionary item
            if bpath != "":
                bkQueryDict[bkFields[i]] = bpath
            if defaultPP:
                # PP was empty, try once more to get the Event Type
                defaultPP = False
            else:
                # Go to next item
                i += 1
            if i == len(bkFields):
                break

        # Set default event type to real data
        if bkQueryDict.get("ConfigName") != "MC" and not bkQueryDict.get("EventType"):
            bkQueryDict["EventType"] = "90000000"
        if bkQueryDict.get("EventType") == "ALL":
            bkQueryDict.pop("EventType")

    return bkQueryDict



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

known_working_groups = [
    "B2CC",
    "B2OC",
    "BandQ",
    "BnoC",
    "Calib",
    "Calo",
    "Charm",
    "DPA",
    "HLT",
    "IFT",
    "Luminosity",
    "PID",
    "QCD",
    "QEE",
    "RD",
    "RTA",
    "Simulation",
    "SL",
    "FlavourTagging",
    "Tracking",
]

known_data_types = ["Upgrade", "2011", "2012", "2015", "2016", "2017", "2018"]

known_input_types = ["DST", "MDST"]

allowed_priorities = ["1a", "1b", "2a", "2b"]



###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

from os.path import dirname, join

import pytest


def test_split_by_level_simple_WARNING():
    from LbAPCommon.validators.logs import _split_by_level

    result = _split_by_level(
        "2020-06-17 06:49:21 UTC Dsp2KmKppip.LoK...WARNING "
        "LoKi::LifetimeFitter:: Error from LoKi::Fitters::ctau_step,reset StatusCode=407"
    )
    assert dict(result) == {
        "WARNING": [
            (
                1,
                None,
                "2020-06-17 06:49:21 UTC",
                "Dsp2KmKppip.LoK...",
                "LoKi::LifetimeFitter:: Error from LoKi::Fitters::ctau_step,reset StatusCode=407",
            )
        ]
    }


def test_split_by_level_simple_INFO():
    from LbAPCommon.validators.logs import _split_by_level

    result = _split_by_level(
        "2020-06-17 08:04:34 UTC L0MuonFromRaw        INFO "
        "L0MuonProcCand Q4     -> nb of banks seen     : 83033"
    )
    assert dict(result) == {
        "INFO": [
            (
                1,
                None,
                "2020-06-17 08:04:34 UTC",
                "L0MuonFromRaw",
                "L0MuonProcCand Q4     -> nb of banks seen     : 83033",
            )
        ]
    }


def test_split_by_level_simple_SUCCESS():
    from LbAPCommon.validators.logs import _split_by_level

    result = _split_by_level(
        "2020-06-17 07:47:04 UTC EventSelector     SUCCESS "
        "Reading Event record 930001. Record number within stream 8: 18066"
    )
    assert dict(result) == {
        "SUCCESS": [
            (
                1,
                None,
                "2020-06-17 07:47:04 UTC",
                "EventSelector",
                "Reading Event record 930001. Record number within stream 8: 18066",
            )
        ]
    }


@pytest.mark.parametrize(
    "log_fn,expected_counts",
    [
        ("error-failed-to-read-file.log", (0, 0, 684, 2, 4, 0, 0, 382)),
        ("error-illegal-instruction.log", (0, 0, 8, 0, 0, 0, 0, 0)),
        ("error-missing-shared-library.log", (0, 0, 0, 0, 0, 0, 0, 0)),
        ("error-platform-unsupported.log", (0, 0, 0, 0, 0, 0, 0, 0)),
        ("good-DaVinci_00110296_00000038_1.log", (0, 0, 276, 3, 0, 0, 0, 345)),
        ("good-DaVinci_00110296_00000194_1.log", (0, 0, 306, 3, 0, 0, 0, 346)),
        ("good-Gauss_00104988_00000011_1.log", (0, 0, 3039, 63, 4, 0, 0, 401)),
    ],
)
def test_split_by_level_full(log_fn, expected_counts):
    from LbAPCommon.validators.logs import _split_by_level

    with open(join(dirname(__file__), "example-logs", log_fn), "rt") as fp:
        text = fp.read()
    result = _split_by_level(text)

    levels = [
        "VERBOSE",
        "DEBUG",
        "INFO",
        "WARNING",
        "ERROR",
        "FATAL",
        "ALWAYS",
        "SUCCESS",
    ]
    counts = tuple(len(result[level]) for level in levels)
    print(counts)
    for level, expected in zip(levels, expected_counts):
        assert len(result[level]) == expected



###############################################################################
# (c) Copyright 2020-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

__all__ = [
    "match_output_filenames",
    "status_from_xml_summary",
    "parse_bookkeeping_xml",
    "count_log_messages",
    "explain_log",
]

import re
from os.path import basename

from lxml import etree

from .bookkeeping_xml import parse_bookkeeping_xml
from .logs import count_log_messages, explain_log


def match_output_filenames(filenames, output_filetypes, application_name, step_index):
    """File the XML, log and output filenames from a test job.

    Parameters
    ----------
    filenames : :obj:`list` of :obj:`str`
        The filenames found in the test output
    output_filetypes : :obj:`list` of :obj:`str`
        The output filetype(s) of the job
    application_name : :obj:`str`
        Name of the application being ran (e.g. DaVinci)
    step_index : :obj:`int`
        The index of this test within the job

    Returns
    -------
    xml_summary_fn : :obj:`str` or None
        The filename of the XML summary
    xml_bk_fn : :obj:`str` or None
        The filename of the Bookkeeping summary
    log_fn : :obj:`str` or None
        The filename of the application log
    output_fns : :obj:`dict`
        A mapping of output filetype to a filename or None
    """
    expected_log_fn = "%s_00012345_00006789_%s.log" % (
        application_name,
        step_index,
    )

    xml_summary_fn = None
    xml_bk_fn = None
    log_fn = None
    output_fns = {ft: None for ft in output_filetypes}
    for fn in filenames:
        fn = basename(fn)
        # Look for the XML summary
        if re.match(r"^summary.*\.xml$", fn):
            if xml_summary_fn is not None:
                raise NotImplementedError(fn, xml_summary_fn)
            xml_summary_fn = fn
        # Look for the Bookkeeping XML
        if re.match(r"^bookkeeping_.*\.xml$", fn):
            if xml_bk_fn is not None:
                raise NotImplementedError(fn, xml_bk_fn)
            xml_bk_fn = fn
        # Look for the application log file
        if expected_log_fn == fn:
            if log_fn is not None:
                raise NotImplementedError(fn, log_fn)
            log_fn = fn
        # Look for the output filenames
        for output_filetype in output_filetypes:
            if fn.upper().endswith(output_filetype):
                if output_fns[output_filetype] is not None:
                    raise NotImplementedError(
                        (
                            "Something is very wrong! Multiple files have been found "
                            "that match {output_filetype} however files in the "
                            "bookkeeping are case-insensitive. The following "
                            "filenames were uploaded by the job: {filenames}"
                        ).format(output_filetype=output_filetype, filenames=filenames)
                    )
                output_fns[output_filetype] = fn

    return xml_summary_fn, xml_bk_fn, log_fn, output_fns


def status_from_xml_summary(fp):
    """Check if the job succeeded using the XML summary file.

    Args:
        fp (file-like): _description_

    Raises:
        NotImplementedError: _description_

    Returns:
        bool: whether the XML summary reports the job as passed
    """
    try:
        xml_summary = etree.parse(fp, base_url="LbAnalysisProductions/data/xml")
        success_value = xml_summary.find(".//success").text.lower()
    except Exception:
        # Be resistant to malformed XML
        return False
    else:
        if success_value == "true":
            return True
        elif success_value == "false":
            return False
        else:
            raise NotImplementedError(success_value)



###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import json
import pickle
import pprint
from argparse import ArgumentParser

from DecayTreeTupleBase.DecayTreeTupleBaseConf import (
    DecayTreeTuple,
    EventTuple,
    MCDecayTreeTuple,
)


def parse_args():
    """Parse command line arguments with `argparse`.

    Returns:
        ArgumentParser-parsed arguments.
    """
    parser = ArgumentParser()
    parser.add_argument("--pkl", required=True, help="Pickle file to read")
    parser.add_argument("--output", required=True, help="Name of output .json file")
    parser.add_argument(
        "--debug", action="store_true", default=False, help="output debug information"
    )
    return parser.parse_args()


def pkl_json_dump(pkl, output, debug=False):
    """Move relevant information from pickle to JSON format.

    Args:
        pkl (str): Pickle file path.
        output (str): Output JSON file path.
        debug (bool, optional): Debug flag. Defaults to False.
    """
    output_dict = {"DecayTreeTuple": [], "MCDecayTreeTuple": [], "EventTuple": []}

    # Open the pickle file
    with open(pkl, "rb") as fp:
        data = pickle.load(fp)

    # Filter the pickle file
    for key in data:
        if isinstance(data[key], DecayTreeTuple):
            output_dict["DecayTreeTuple"].append(key)
        elif isinstance(data[key], MCDecayTreeTuple):
            output_dict["MCDecayTreeTuple"].append(key)
        elif isinstance(data[key], EventTuple):
            output_dict["EventTuple"].append(key)

    if debug:
        print("\nOutput for the .json file:")
        pprint.pprint(output_dict)

    # Save useful information in JSON file
    with open(output, "wt") as f:
        json.dump(output_dict, f)


if __name__ == "__main__":

    args = parse_args()

    pkl_json_dump(args.pkl, args.output, debug=args.debug)



###############################################################################
# (c) Copyright 2021-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

import json

import uproot

from LbAPCommon.parsing import is_simulation_job


def validate_options(json_file: str, ntuple_fn: str, job_name: str, prod_data: dict):
    """Validate YAML options.

    Check existence of expected TTrees and ensure at least one
    DecayTree and MCDecayTree (for MC samples) exist(s) in output files.

    Args:
        json_file (str): json_file listing the expected TTrees
        ntuple_fn (str): Local test output tuple to validate against
        job_name (str): Name of job to validate
        prod_data (dict): Entire production information from yaml parsing

    Returns:
        tuple[list[str]]: Errors if any expected TTrees are not found in job output,
            warnings if at least one (MC)DecayTree isn't found in job output
    """
    with open(json_file, "rb") as fp:
        json_dump = json.load(fp)

    file = uproot.open(ntuple_fn)
    tuple_keys = file.keys(cycle=False)

    errors = []
    warnings = []

    mc_decay_tree = any({"MCDecayTree" in key for key in tuple_keys})
    decay_tree = any(
        {("DecayTree" in key and "MCDecayTree" not in key) for key in tuple_keys}
    )

    if not decay_tree:
        warnings.append(
            "No DecayTree detected in the output file! Is this intentional?"
        )

    # It's not possible to statically check if using transform_ids as input
    if "transform_ids" not in prod_data[job_name]["input"]:
        if is_simulation_job(prod_data, job_name) and not mc_decay_tree:
            warnings.append(
                "No MCDecayTree detected in the output file! Is this intentional?"
            )

    # Checking DecayTreeTuples
    for ntuple in json_dump["DecayTreeTuple"]:
        if ntuple not in tuple_keys:
            errors.append(f"ERROR: DecayTreeTuple {ntuple} missing in test result")

    # Checking MCDecayTreeTuples
    for ntuple in json_dump["MCDecayTreeTuple"]:
        if ntuple not in tuple_keys:
            errors.append(f"ERROR: MCDecayTreeTuple {ntuple} missing in test result")

    # Checking MCDecayTreeTuples
    for ntuple in json_dump["EventTuple"]:
        if ntuple not in tuple_keys:
            errors.append(f"ERROR: EventTuple {ntuple} missing in test result")

    return errors, warnings



###############################################################################
# (c) Copyright 2021 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################
import re
from pathlib import Path

import pytest

from LbAPCommon import checks

pytest.importorskip("XRootD")


def test_num_entries_passing():
    data_in = {
        "check": {
            "type": "num_entries",
            "count": 1000,
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert result.passed
    assert result.messages == ["Found 5135823 in DecayTree (1000 required)"]
    assert result.tree_data["DecayTree"]["num_entries"] == 5135823


def test_num_entries_passing_multiple_files():
    data_in = {
        "check": {
            "type": "num_entries",
            "count": 1000,
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root",
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetUp.root",
        ],
    )["check"]
    assert result.passed
    assert result.messages == ["Found 8556118 in DecayTree (1000 required)"]
    assert result.tree_data["DecayTree"]["num_entries"] == 8556118


def test_num_entries_failing():
    data_in = {
        "check": {
            "type": "num_entries",
            "count": 1000000000,
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == ["Found 5135823 in DecayTree (1000000000 required)"]
    assert result.tree_data["DecayTree"]["num_entries"] == 5135823


def test_num_entries_failing_tree_name():
    data_in = {
        "check": {
            "type": "num_entries",
            "count": 1000,
            "tree_pattern": "RandomName",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == ["No TTree objects found that match RandomName"]
    for _key, data in result.tree_data.items():
        assert data["histograms"] == []


def test_num_entries_per_invpb_passing():
    data_in = {
        "check": {
            "type": "num_entries_per_invpb",
            "count_per_invpb": 10000,
            "tree_pattern": "DecayTree",
            "lumi_pattern": "LumiTuple",
        }
    }
    file_list = [Path(__file__).parent.absolute() / "example_tuple_with_lumi.root"]
    result = checks.run_job_checks(
        ["check"],
        data_in,
        file_list,
    )["check"]
    assert result.passed
    assert result.messages == [
        "Found 23026.2 entries per unit luminosity (pb-1) in DecayTree (10000 required)"
    ]
    assert result.tree_data["DecayTree"]["num_entries_per_invpb"] == 23026.2


def test_num_entries_per_invpb_failing():
    data_in = {
        "check": {
            "type": "num_entries_per_invpb",
            "count_per_invpb": 10000000,
            "tree_pattern": "DecayTree",
            "lumi_pattern": "LumiTuple",
        }
    }
    file_list = [Path(__file__).parent.absolute() / "example_tuple_with_lumi.root"]
    result = checks.run_job_checks(
        ["check"],
        data_in,
        file_list,
    )["check"]
    assert not result.passed
    assert result.messages == [
        "Found 23026.2 entries per unit luminosity (pb-1) in DecayTree (10000000 required)"
    ]
    assert result.tree_data["DecayTree"]["num_entries_per_invpb"] == 23026.2


def test_num_entries_per_invpb_failing_MC():
    data_in = {
        "check": {
            "type": "num_entries_per_invpb",
            "count_per_invpb": 10000,
            "tree_pattern": "DecayTree",
            "lumi_pattern": "LumiTuple",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == [
        "Failed to get luminosity information (total luminosity = 0)"
    ]


def test_num_entries_per_invpb_failing_MC_nameTTree():
    data_in = {
        "check": {
            "type": "num_entries_per_invpb",
            "count_per_invpb": 10000,
            "tree_pattern": "RandomName",
            "lumi_pattern": "LumiTuple",
        }
    }
    file_list = [Path(__file__).parent.absolute() / "example_tuple_with_lumi.root"]
    result = checks.run_job_checks(
        ["check"],
        data_in,
        file_list,
    )["check"]
    assert not result.passed
    assert result.messages == [
        "No TTree objects found that match RandomName",
    ]


def test_range_check_passing():
    data_in = {
        "check": {
            "type": "range",
            "expression": "H1_PZ",
            "limits": {"min": 0.0, "max": 500000.0},
            "n_bins": 50,
            "blind_ranges": [],
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert result.passed
    assert result.messages == [
        "Histogram of H1_PZ successfully filled from TTree DecayTree (contains 5134459.0 events)"
    ]
    assert list(result.tree_data.keys()) == ["DecayTree"]
    assert not result.tree_data["DecayTree"]["histograms"] == []
    assert result.tree_data["DecayTree"]["num_entries"] == 5134459


def test_range_check_passing_multiple_files():
    data_in = {
        "check": {
            "type": "range",
            "expression": "H1_PZ",
            "limits": {"min": 0.0, "max": 500000.0},
            "n_bins": 50,
            "blind_ranges": [],
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root",
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetUp.root",
        ],
    )["check"]
    assert result.passed
    assert result.messages == [
        "Histogram of H1_PZ successfully filled from TTree DecayTree (contains 8553802.0 events)"
    ]
    assert list(result.tree_data.keys()) == ["DecayTree"]
    assert not result.tree_data["DecayTree"]["histograms"] == []
    assert result.tree_data["DecayTree"]["num_entries"] == 8553802


def test_range_check_failing_range():
    data_in = {
        "check": {
            "type": "range",
            "expression": "H1_PZ",
            "limits": {"min": -100000.0, "max": -99999.0},
            "n_bins": 50,
            "blind_ranges": [],
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == ["No events found in range for Tree DecayTree"]
    assert list(result.tree_data.keys()) == ["DecayTree"]
    for h in result.tree_data["DecayTree"]["histograms"]:
        assert h.sum() == 0


def test_range_check_failing_missing_branch():
    data_in = {
        "check": {
            "type": "range",
            "expression": "Dst_M",
            "limits": {"min": 1800.0, "max": 2300.0},
            "n_bins": 50,
            "blind_ranges": [],
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    message = result.messages[0]
    pattern = r"Missing branch in "
    matched = False
    if re.match(pattern, message):
        matched = True
    assert not result.passed
    assert matched
    assert list(result.tree_data.keys()) == ["DecayTree"]
    for h in result.tree_data["DecayTree"]["histograms"]:
        assert h.sum() == 0


def test_range_check_failing_tree_name():
    data_in = {
        "check": {
            "type": "range",
            "expression": "H1_PZ",
            "limits": {"min": 0.0, "max": 500000.0},
            "n_bins": 50,
            "blind_ranges": [],
            "tree_pattern": "RandomName",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == ["No TTree objects found that match RandomName"]
    assert list(result.tree_data.keys()) == []


def test_range_check_failing_bad_mean():
    data_in = {
        "check": {
            "type": "range",
            "expression": "H1_PZ",
            "limits": {"min": 0.0, "max": 500000.0},
            "n_bins": 50,
            "blind_ranges": [],
            "tree_pattern": "DecayTree",
            "exp_mean": 500000.0,
            "mean_tolerance": 1.0,
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == [
        "The observed mean (49110.65313794501) differs from the expected value by 450889.346862055 (<=1.0 required)"
    ]
    assert list(result.tree_data.keys()) == ["DecayTree"]
    assert not result.tree_data["DecayTree"]["histograms"] == []
    assert result.tree_data["DecayTree"]["num_entries"] == 5134459


def test_range_check_failing_bad_stddev():
    data_in = {
        "check": {
            "type": "range",
            "expression": "H1_PZ",
            "limits": {"min": 0.0, "max": 500000.0},
            "n_bins": 50,
            "blind_ranges": [],
            "tree_pattern": "DecayTree",
            "exp_std": 500000.0,
            "std_tolerance": 1.0,
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == [
        "The observed standard deviation (53099.76473607609) differs from the expected value by 446900.2352639239 (<=1.0 required)"
    ]
    assert list(result.tree_data.keys()) == ["DecayTree"]
    assert not result.tree_data["DecayTree"]["histograms"] == []
    assert result.tree_data["DecayTree"]["num_entries"] == 5134459


def test_range_check_nd_passing():
    data_in = {
        "check": {
            "type": "range_nd",
            "expressions": {"x": "H1_PZ", "y": "H2_PZ", "z": "H2_PX"},
            "limits": {
                "x": {"min": 0.0, "max": 500000.0},
                "y": {"min": 0.0, "max": 500000.0},
                "z": {"min": 0.0, "max": 500000.0},
            },
            "n_bins": {
                "x": 50,
                "y": 50,
                "z": 50,
            },
            "blind_ranges": [],
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert result.passed
    assert result.messages == [
        "Histogram of H1_PZ, H2_PZ successfully filled from TTree DecayTree (contains 5134453.0 events)",
        "Histogram of H1_PZ, H2_PX successfully filled from TTree DecayTree (contains 2525193.0 events)",
        "Histogram of H2_PZ, H2_PX successfully filled from TTree DecayTree (contains 2525806.0 events)",
        "Histogram of H1_PZ, H2_PZ, H2_PX successfully filled from TTree DecayTree (contains 2525189.0 events)",
    ]
    assert list(result.tree_data.keys()) == ["DecayTree"]
    assert not result.tree_data["DecayTree"]["histograms"] == []


def test_range_check_nd_failing_num_var():
    data_in = {
        "check": {
            "type": "range_nd",
            "expressions": {"x": "H1_PZ"},
            "limits": {
                "x": {"min": 0.0, "max": 500000.0},
            },
            "n_bins": {
                "x": 50,
                "y": 50,
            },
            "blind_ranges": [],
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == ["Expected two or three variables."]
    assert list(result.tree_data.keys()) == []


def test_range_check_nd_failing_missing_limit():
    data_in = {
        "check": {
            "type": "range_nd",
            "expressions": {"x": "H1_PZ", "y": "H2_PZ"},
            "limits": {
                "x": {"min": 0.0, "max": 500000.0},
            },
            "n_bins": {
                "x": 50,
                "y": 50,
            },
            "blind_ranges": [],
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == [
        "For each variable, a corresponding range should be defined."
    ]
    assert list(result.tree_data.keys()) == []


def test_range_check_nd_failing_missing_branch():
    data_in = {
        "check": {
            "type": "range_nd",
            "expressions": {"x": "H1_PZ", "y": "Dst_M-D0_M"},
            "limits": {
                "x": {"min": 0.0, "max": 500000.0},
                "y": {"min": 138.0, "max": 150.0},
            },
            "n_bins": {
                "x": 50,
                "y": 50,
            },
            "blind_ranges": [],
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    message = result.messages[0]
    pattern = r"Missing branch in "
    matched = False
    if re.match(pattern, message):
        matched = True
    assert not result.passed
    assert matched
    assert list(result.tree_data.keys()) == ["DecayTree"]
    for h in result.tree_data["DecayTree"]["histograms"]:
        assert h.sum() == 0


def test_range_check_nd_failing_range():
    data_in = {
        "check": {
            "type": "range_nd",
            "expressions": {"x": "H1_PZ", "y": "H2_PZ"},
            "limits": {
                "x": {"min": -1000000.0, "max": -999999.0},
                "y": {"min": -1000000.0, "max": -999999.0},
            },
            "n_bins": {
                "x": 50,
                "y": 50,
            },
            "blind_ranges": [],
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == ["No events found in range for Tree DecayTree"]
    assert list(result.tree_data.keys()) == ["DecayTree"]
    for h in result.tree_data["DecayTree"]["histograms"]:
        assert h.sum() == 0


def test_range_check_nd_failing_tree_name():
    data_in = {
        "check": {
            "type": "range_nd",
            "expressions": {"x": "H1_PZ", "y": "H2_PZ"},
            "limits": {
                "x": {"min": 0.0, "max": 500000.0},
                "y": {"min": 0.0, "max": 500000.0},
            },
            "n_bins": {
                "x": 50,
                "y": 50,
            },
            "blind_ranges": [],
            "tree_pattern": "RandomName",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == ["No TTree objects found that match RandomName"]
    assert list(result.tree_data.keys()) == []


def test_range_check_bkg_subtracted_passing():
    data_in = {
        "check": {
            "type": "range_bkg_subtracted",
            "expression": "D0_PT",
            "limits": {"min": 0.0, "max": 500000.0},
            "expr_for_subtraction": "D0_MM",
            "mean_sig": 1865.0,
            "background_shift": 30.0,
            "background_window": 10.0,
            "signal_window": 20.0,
            "n_bins": 50,
            "blind_ranges": {"min": 10000.0, "max": 30000.0},
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/MasterclassDatasets/D0lifetime/2014/MasterclassData.root"
        ],
    )["check"]
    assert result.passed
    assert (
        result.messages[0]
        == "Background subtraction performed successfully for Tree DecayTree"
    )
    assert list(result.tree_data.keys())[0] == "DecayTree"
    assert not result.tree_data["DecayTree"]["histograms"] == []


def test_range_check_bkg_subtracted_failing_range():
    data_in = {
        "check": {
            "type": "range_bkg_subtracted",
            "expression": "D0_PT",
            "limits": {"min": -1000000.0, "max": -999999.0},
            "expr_for_subtraction": "D0_MM",
            "mean_sig": 1865.0,
            "background_shift": 30.0,
            "background_window": 10.0,
            "signal_window": 20.0,
            "n_bins": 50,
            "blind_ranges": {"min": 10000.0, "max": 30000.0},
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/MasterclassDatasets/D0lifetime/2014/MasterclassData.root"
        ],
    )["check"]
    assert not result.passed
    assert (
        result.messages[0]
        == "Not enough events for background subtraction found in range for Tree DecayTree"
    )
    assert list(result.tree_data.keys())[0] == "DecayTree"


def test_range_check_bkg_subtracted_failing_missing_branch():
    data_in = {
        "check": {
            "type": "range_bkg_subtracted",
            "expression": "D0_PT",
            "limits": {"min": 0.0, "max": 500000.0},
            "expr_for_subtraction": "Dst_M",
            "mean_sig": 1865.0,
            "background_shift": 30.0,
            "background_window": 10.0,
            "signal_window": 20.0,
            "n_bins": 50,
            "blind_ranges": {"min": 10000.0, "max": 30000.0},
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/MasterclassDatasets/D0lifetime/2014/MasterclassData.root"
        ],
    )["check"]
    message = result.messages[0]
    print(message)
    pattern = r"Missing branch in "
    matched = False
    if re.match(pattern, message):
        matched = True
    assert not result.passed
    assert matched
    assert list(result.tree_data.keys())[0] == "DecayTree"


def test_range_check_bkg_subtracted_failing_tree_name():
    data_in = {
        "check": {
            "type": "range_bkg_subtracted",
            "expression": "D0_PT",
            "limits": {"min": 0.0, "max": 500000.0},
            "expr_for_subtraction": "D0_MM",
            "mean_sig": 1865.0,
            "background_shift": 30.0,
            "background_window": 10.0,
            "signal_window": 20.0,
            "n_bins": 50,
            "blind_ranges": {"min": 10000.0, "max": 30000.0},
            "tree_pattern": "RandomName",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/MasterclassDatasets/D0lifetime/2014/MasterclassData.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == ["No TTree objects found that match RandomName"]
    assert list(result.tree_data.keys()) == []


def test_branches_exist_passing():
    data_in = {
        "check": {
            "type": "branches_exist",
            "branches": ["H1_PZ", "H2_PZ"],
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert result.passed
    assert result.messages == ["All required branches were found in Tree DecayTree"]


def test_branches_exist_failing_missing_branches():
    data_in = {
        "check": {
            "type": "branches_exist",
            "branches": ["H1_PZ", "H2_PZ", "RandomName"],
            "tree_pattern": "DecayTree",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == [
        "Required branches not found in Tree DecayTree: ['RandomName']"
    ]


def test_branches_exist_failing_tree_name():
    data_in = {
        "check": {
            "type": "branches_exist",
            "branches": ["H1_PZ", "H2_PZ"],
            "tree_pattern": "RandomName",
        }
    }
    result = checks.run_job_checks(
        ["check"],
        data_in,
        [
            "root://eospublic.cern.ch//eos/opendata/lhcb/AntimatterMatters2017/data/B2HHH_MagnetDown.root"
        ],
    )["check"]
    assert not result.passed
    assert result.messages == ["No TTree objects found that match RandomName"]



###############################################################################
# (c) Copyright 2020-2022 CERN for the benefit of the LHCb Collaboration      #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################


def lint_all(data):
    """Placeholder for linting.

    Args:
        data: ...

    Raises:
        NotImplementedError: This function is not implemented.
    """
    raise NotImplementedError()



###############################################################################
# (c) Copyright 2020 CERN for the benefit of the LHCb Collaboration           #
#                                                                             #
# This software is distributed under the terms of the GNU General Public      #
# Licence version 3 (GPL Version 3), copied verbatim in the file "COPYING".   #
#                                                                             #
# In applying this licence, CERN does not waive the privileges and immunities #
# granted to it by virtue of its status as an Intergovernmental Organization  #
# or submit itself to any jurisdiction.                                       #
###############################################################################

from os.path import dirname, join

from LbAPCommon.validators import explain_log


def _load_log(filename):
    with open(join(dirname(__file__), "example-logs", filename), "rt") as fp:
        return fp.read()


def test_missing_shared_library():
    text = _load_log("error-missing-shared-library.log")
    explanations, suggestions, errors = explain_log(text)

    assert len(errors) == 1
    error_message, line_numbers = errors.pop()
    assert "HEP_OSlibs" in error_message
    assert line_numbers is None


def test_illegal_instruction():
    text = _load_log("error-illegal-instruction.log")
    explanations, suggestions, errors = explain_log(text)

    assert len(errors) == 1
    error_message, line_numbers = errors.pop()
    assert "compiled with instructions" in error_message
    assert line_numbers is None


def test_platform_unsupported():
    text = _load_log("error-platform-unsupported.log")
    explanations, suggestions, errors = explain_log(text)

    assert len(errors) == 1
    error_message, line_numbers = errors.pop()
    assert "lb-describe-platform" in error_message
    assert line_numbers is None


def test_related_info_missing():
    text = _load_log("error-related-info-missing.log")
    explanations, suggestions, errors = explain_log(text)

    # Errors
    (relinfo_error,) = sorted(errors)

    error_message, line_numbers = relinfo_error
    assert "RelatedInfo" in error_message
    assert len(line_numbers) == 54

    # Suggestions
    (cal_mc_suggest,) = sorted(suggestions)

    error_message, line_numbers = cal_mc_suggest
    assert "calorimeter MC-truth" in error_message
    assert len(line_numbers) == 9

    # Suggestions
    (histogram_not_set,) = sorted(explanations)

    error_message, line_numbers = histogram_not_set
    assert "Histograms are not being saved" in error_message
    assert "harmless" in error_message
    assert line_numbers == [4678]


def test_failed_to_read_file():
    text = _load_log("error-failed-to-read-file.log")
    explanations, suggestions, errors = explain_log(text)

    assert len(errors) == 1
    error_message, line_numbers = errors.pop()
    assert "error accessing the input" in error_message
    assert line_numbers == [1085]


def test_good_davinci_38():
    text = _load_log("good-DaVinci_00110296_00000038_1.log")
    explanations, suggestions, errors = explain_log(text)

    assert errors == []


def test_good_davinci_194():
    text = _load_log("good-DaVinci_00110296_00000194_1.log")
    explanations, suggestions, errors = explain_log(text)

    assert errors == []


def test_good_gauss_11():
    text = _load_log("good-Gauss_00104988_00000011_1.log")
    explanations, suggestions, errors = explain_log(text)

    assert errors == []
